import asyncio

import attrs
import logging
import numpy as np
import pyvista as pv
import pyvistaqt as pvqt
from qtpy import QtWidgets, QtGui, QtCore
import typing as tp

from NaviNIBS.Navigator.Model.Session import Session
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.GUI.QueuedRedrawMixin import QueuedRedrawMixin
from NaviNIBS.util.numpy import array_equalish
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.Transforms import composeTransform, applyTransform, applyDirectionTransform
from NaviNIBS.util.pyvista import DefaultBackgroundPlotter, RemotePlotterProxy, setActorUserTransform, Actor

logger = logging.getLogger(__name__)


@attrs.define
class MRISliceView(QueuedRedrawMixin):
    _normal: tp.Union[str, np.ndarray] = 'x'  # if an ndarray, should actually be 3x3 transform matrix from view pos to world space, not just 3-elem normal direction
    _label: tp.Optional[str] = None  # if none, will be labelled according to normal; this assumes normal won't change
    _cameraOffsetDist: float = 100  # distance from slice to camera
    _session: tp.Optional[Session] = attrs.field(default=None, repr=False)
    _sliceOrigin: tp.Optional[np.ndarray] = None

    _slicePlotMethod: str = 'cameraClippedVolume'
    _doShowScalarBar: bool = False

    _plotter: DefaultBackgroundPlotter = attrs.field(init=False, default=None)
    _plotterInitialized: bool = attrs.field(init=False, default=False)
    _plotterPickerInitialized: bool = attrs.field(init=False, default=False)
    _lineActors: tp.Dict[str, pv.Line] = attrs.field(init=False, factory=dict)
    _volActor: Actor | None = attrs.field(init=False, factory=dict)

    _backgroundColor: str = '#000000'
    _opacity: float = 0.5

    finishedAsyncInit: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    sigSliceOriginChanged: Signal = attrs.field(init=False, factory=Signal)
    sigNormalChanged: Signal = attrs.field(init=False, factory=Signal)
    sigSliceTransformChanged: Signal = attrs.field(init=False, factory=Signal)

    def __attrs_post_init__(self):
        QueuedRedrawMixin.__attrs_post_init__(self)

        self.sigSliceTransformChanged.connect(self.updateView)
        if self._session is not None:
            self._session.MRI.sigDataChanged.connect(self._onMRIDataChanged)

        if self._plotter is None:
            self._plotter = DefaultBackgroundPlotter()
        else:
            pass  # presumably was initialized by subclass

        asyncio.create_task(asyncTryAndLogExceptionOnError(self._finish_init))

    async def _finish_init(self):
        if isinstance(self._plotter, RemotePlotterProxy):
            await self._plotter.isReadyEvent.wait()

        with self._plotter.allowNonblockingCalls():
            self._plotter.set_background(self._backgroundColor)
            _ = self.plotter.camera  # get camera to get past BasePlotter's reset_camera call

        self.finishedAsyncInit.set()

        self.updateView()

    @property
    def label(self):
        if self._label is None:
            assert isinstance(self._normal, str)
            return self._normal
        else:
            return self._label

    @property
    def wdgt(self):
        return self._plotter

    @property
    def plotter(self):
        return self._plotter

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, newSession: tp.Optional[Session]):
        if newSession is self._session:
            return

        if self._session is not None:
            self._session.MRI.sigDataChanged.disconnect(self._onMRIDataChanged)
            self._session.MRI.sigClimChanged.disconnect(self._onMRIClimChanged)

        self._session = newSession
        if self._session is not None:
            self._session.MRI.sigDataChanged.connect(self._onMRIDataChanged)
            self._session.MRI.sigClimChanged.connect(self._onMRIClimChanged)

        self.updateView()

    @property
    def sliceOrigin(self):
        return self._sliceOrigin

    @sliceOrigin.setter
    def sliceOrigin(self, newVal: tp.Optional[np.ndarray]):
        if array_equalish(newVal, self._sliceOrigin):
            # no change
            return
        logger.debug(f'Slice origin changed from {self._sliceOrigin} to {newVal}')
        self._sliceOrigin = newVal
        self.sigSliceOriginChanged.emit()
        self.sigSliceTransformChanged.emit()

    @property
    def sliceTransform(self):
        transf = np.eye(4)
        if isinstance(self._normal, str):
            # TODO: double check these
            if self._normal == 'x':
                transf[0:3, 0:3] = np.asarray([[0, 0, 1], [0, 1, 0], [-1, 0, 0]])
            elif self._normal == 'y':
                transf[0:3, 0:3] = np.asarray([[1, 0, 0], [0, 0, -1], [0, 1, 0]])
            elif self._normal == 'z':
                pass
            else:
                raise NotImplementedError()
        else:
            transf[0:3, 0:3] = self._normal

        transf[0:3, 3] = self._sliceOrigin

        return transf

    @sliceTransform.setter
    def sliceTransform(self, newTransf: np.ndarray):
        newOrigin = newTransf[0:3, 3]
        newRot = newTransf[0:3, 0:3]
        assert array_equalish(newTransf[3, :], np.asarray([0, 0, 0, 1]))

        originChanged = not array_equalish(newOrigin, self._sliceOrigin)
        rotChanged = not array_equalish(self._normal, newRot)

        if not originChanged and not rotChanged:
            return

        self._sliceOrigin = newOrigin
        self._normal = newRot

        self.sigSliceTransformChanged.emit()
        if originChanged:
            self.sigSliceOriginChanged.emit()
        if rotChanged:
            self.sigNormalChanged.emit()

    def _onSlicePointChanged(self):
        pos = self.plotter.picked_point
        logger.debug('Slice point changed: {} {}'.format(self.label, pos))
        if True:
            # ignore any out-of-plane offset that can be caused by slice rendering differences
            if isinstance(self._normal, str):
                pos['xyz'.index(self._normal)] = self._sliceOrigin['xyz'.index(self._normal)]
            else:
                normalDir = self._normal @ np.asarray([0, 0, 1])
                pos = pos - np.dot(pos - self._sliceOrigin, normalDir) * normalDir
        self.sliceOrigin = pos

    def _onSliceScrolled(self, change: int):
        logger.debug('Slice scrolled: {} {}'.format(self.label, change))
        offset = np.zeros((3,))
        if isinstance(self._normal, str):
            offset['xyz'.index(self._normal)] = change
        else:
            offset = self._normal @ np.asarray([0, 0, 1]) * change
        pos = self._sliceOrigin + offset
        self.sliceOrigin = pos

    def _onMouseEvent(self, _, event):
        # TODO: check if `obj` has any info about scroll speed to scroll with larger increment

        if event == 'MouseWheelForwardEvent':
            logger.debug('MouseWheelForwardEvent')
            self._onSliceScrolled(change=1)

        elif event == 'MouseWheelBackwardEvent':
            logger.debug('MouseWheelBackwardEvent')
            self._onSliceScrolled(change=-1)
            
    def _onMRIDataChanged(self):
        if self._plotterInitialized:
            self._clearPlot()
        self.updateView()

    def _onMRIClimChanged(self, dim: str):
        if dim == '2D':
            self._queueRedraw(which='updateClim')

    def _updateClim(self):
        if not self.finishedAsyncInit.is_set():
            # plotter not available yet
            return

        if not self._plotterInitialized:
            # initial plot not created yet
            return

        if self.session is not None and self.session.MRI.isSet:
            with self._plotter.allowNonblockingCalls():
                self._plotter.updateScalarBarRangeWithVol(
                    clim=self.session.MRI.clim2D,
                    volumeKey='MRI',
                    volume=self.session.MRI.dataAsUniformGrid
                )
                self._plotter.render()

    def _clearPlot(self):
        logger.debug('Clearing plot for {} slice'.format(self.label))
        with self._plotter.allowNonblockingCalls():
            self._plotter.clear()
        self.sliceOrigin = None
        self._plotterInitialized = False

    def _redraw(self, which: tp.Union[tp.Optional[str], tp.List[str]] = None, **kwargs):
        super()._redraw(which=which, **kwargs)

        if which is None:
            which = 'all'
            self._redraw(which=which, **kwargs)
            return

        if not isinstance(which, str):
            for subWhich in which:
                self._redraw(which=subWhich, **kwargs)
            return

        match which:
            case 'all':
                self._redraw(which=['updateView'])
                return

            case 'updateView':
                self._updateView()

            case 'updateClim':
                self._updateClim()

            case _:
                raise NotImplementedError

    def updateView(self):
        self._queueRedraw(which='updateView')

    def _updateView(self):
        if self.session is None or self.session.MRI.data is None:
            # no data available
            if self._plotterInitialized:
                self._clearPlot()
            return

        # data available, update display
        logger.debug(f'Updating plot for {self.label} slice (slicePlotMethod={self._slicePlotMethod})')

        if self._sliceOrigin is None:
            self.sliceOrigin = (self.session.MRI.data.affine @ np.append(np.asarray(self.session.MRI.data.shape)/2, 1))[:-1]
            return  # prev line will have triggered its own update

        if not self.finishedAsyncInit.is_set():
            # plotter not available yet
            return

        if not self._plotterPickerInitialized:
            logger.debug('Initializing plot for {} slice'.format(self.label))
            with self.plotter.allowNonblockingCalls():
                self.plotter.enable_parallel_projection()
                self.plotter.enable_point_picking(left_clicking=True,
                                                   show_message=False,
                                                   show_point=False,
                                                   pickable_window=True,
                                                   callback=lambda newPt: self._onSlicePointChanged())
                self.plotter.enable_image_style()
                for event in ('MouseWheelForwardEvent', 'MouseWheelBackwardEvent'):
                    self.plotter.addIrenStyleClassObserver(
                        event=event,
                        callback=lambda obj, event: self._onMouseEvent(obj,event))
            self._plotterPickerInitialized = True

        if self._slicePlotMethod == 'slicedSurface':
            # single-slice plot
            slice = self.session.MRI.dataAsUniformGrid.slice(
                normal=applyDirectionTransform(self.session.MRI.scannerToDataTransf, self._normal),
                origin=applyTransform(self.session.MRI.scannerToDataTransf, self._sliceOrigin))
            # this slicing is very slow for some reason
            with self._plotter.allowNonblockingCalls():
                self._plotter.add_mesh(slice,
                                       name='slice',
                                       cmap='gray',
                                       render=False,
                                       reset_camera=False,
                                       show_scalar_bar=self._doShowScalarBar)
                if isinstance(self._normal, str):
                    self.plotter.camera_position = 'xyz'.replace(self._normal, '')
                else:
                    raise NotImplementedError()  # TODO

        elif self._slicePlotMethod == 'cameraClippedVolume':
            # volume plotting with camera clipping
            if not self._plotterInitialized:
                logger.debug('Getting MRI data as uniform grid')
                vol = self.session.MRI.dataAsUniformGrid
                logger.debug('Initializing volume plot of data')
                self._volActor = self._plotter.add_volume(vol,
                                         scalars='MRI',
                                         name='MRI',
                                         mapper='gpu',
                                         clim=self.session.MRI.clim2D,
                                         scalar_bar_args=dict(
                                             title='',
                                             color='white',
                                             vertical=True,
                                             position_x=0.02,
                                             position_y=0.55,
                                             height=0.4,
                                             label_font_size=12
                                         ),
                                         opacity=[0, self._opacity, self._opacity],
                                         cmap='gray',
                                         show_scalar_bar=self._doShowScalarBar,
                                         render=False,
                                         reset_camera=False)
                with self._plotter.allowNonblockingCalls():
                    setActorUserTransform(self._volActor, self.session.MRI.dataToScannerTransf)

        logger.debug('Setting crosshairs for {} plot'.format(self.label))
        lineLength = 300  # TODO: scale by image size
        if isinstance(self._normal, str):
            crosshairAxes = 'xyz'.replace(self._normal, '')
        else:
            crosshairAxes = 'xy'  # will be transformed below
        centerGapLength = 10  # TODO: scale by image size

        offsetDir = np.zeros((3,))
        if isinstance(self._normal, str):
            offsetDir['xyz'.index(self._normal)] = 1
            if self._normal == 'y':
                offsetDir *= -1  # reverse direction for coronal slice to match L/R of other views
        else:
            offsetDir = self._normal @ np.asarray([0, 0, 1])  # TODO: double check

        for axis in crosshairAxes:
            mask = np.zeros((1, 3))
            mask[0, 'xyz'.index(axis)] = 1
            for iDir, dir in enumerate((-1, 0, 1)):
                if dir == 0:
                    pts = np.asarray([centerGapLength / 2, -centerGapLength / 2])[:, np.newaxis] * mask
                    width = 1
                else:
                    pts = dir * np.asarray([centerGapLength / 2, lineLength])[:, np.newaxis] * mask
                    width = 2

                if isinstance(self._normal, str):
                    pts += self._sliceOrigin
                else:
                    viewToWorldTransf = composeTransform(self._normal, self._sliceOrigin)
                    pts = applyTransform(viewToWorldTransf, pts)

                if True:
                    # add z offset to make sure crosshairs appear above other actors
                    pts -= offsetDir * (self._cameraOffsetDist * 0.02)

                lineKey = 'Crosshair_{}_{}_{}'.format(self.label, axis, iDir)
                if not self._plotterInitialized:
                    line = self._plotter.add_lines(pts, color='#11DD11', width=width, name=lineKey)
                    with self._plotter.allowNonblockingCalls():
                        line.SetUseBounds(False)  # don't include for determining camera zoom, etc.
                    self._lineActors[lineKey] = line
                else:
                    with self._plotter.allowNonblockingCalls():
                        logger.debug('Moving previous crosshairs')
                        line = self._lineActors[lineKey]
                        pts_pv = pv.lines_from_points(pts)
                        line.GetMapper().SetInputData(pts_pv)


        if True:
            with self._plotter.allowNonblockingCalls():
                self._plotter.camera.position = offsetDir * self._cameraOffsetDist + self._sliceOrigin
        else:
            # hack to prevent resetting clipping range due to pyvista implementation quirk
            tmp = self._plotter.camera._renderer
            self._plotter.camera._renderer = None
            self._plotter.camera.position = offsetDir * 100 + self._sliceOrigin
            self._plotter.camera._renderer = tmp

        with self.plotter.allowNonblockingCalls():
            self.plotter.camera.focal_point = self._sliceOrigin
            if isinstance(self._normal, str):
                upDir = np.roll(offsetDir, (2, 1, 2)['xyz'.index(self._normal)])
                if self._normal == 'y':
                    upDir *= -1
            else:
                upDir = self._normal @ np.asarray([0, 1, 0])
            self.plotter.camera.up = upDir
            if self._slicePlotMethod == 'cameraClippedVolume':
                self.plotter.camera.clipping_range = (self._cameraOffsetDist * 0.98, self._cameraOffsetDist * 1.02)

            self.plotter.camera.parallel_scale = self._cameraOffsetDist

            self._plotterInitialized = True

            self.plotter.render()


@attrs.define
class MRI3DView(MRISliceView):
    _clim: tp.Tuple[float, float] = (300, 1000)  # TODO: set to auto-initialize instead of hardcoding default

    def __attrs_post_init__(self):
        MRISliceView.__attrs_post_init__(self)

    @property
    def label(self):
        if self._label is None:
            return 'MRI3D'
        else:
            return self._label

    def _updateView(self):

        if self.session is None or self.session.MRI.data is None:
            # no data available
            if self._plotterInitialized:
                logger.debug('Clearing plot for {} slice'.format(self.label))
                self._plotter.clear()

                self.sliceOrigin = None
                self._plotterInitialized = False
            return

        # data available, update display
        logger.debug('Updating plot for {} slice'.format(self.label))
        if self._sliceOrigin is None:
            self.sliceOrigin = (self.session.MRI.data.affine @ np.append(np.asarray(self.session.MRI.data.shape) / 2,
                                                                         1))[:-1]
            return  # prev line will have triggered its own update

        if not self.finishedAsyncInit.is_set():
            # plotter not ready
            return

        if not self._plotterInitialized:
            logger.debug('Initializing 3D plot')

            self._volActor = self._plotter.add_volume(self.session.MRI.dataAsUniformGrid.gaussian_smooth(),
                                                scalars='MRI',
                                                scalar_bar_args=dict(
                                                    title='',
                                                    color='white',
                                                    vertical=True,
                                                    position_x=0.02,
                                                    position_y=0.55,
                                                    height=0.4,
                                                    label_font_size=12
                                                ),
                                                name='vol',
                                                clim=self.session.MRI.clim3D,
                                                cmap='gray',
                                                show_scalar_bar=self._doShowScalarBar,
                                                mapper='gpu',
                                                opacity=[0, self._opacity, self._opacity],
                                                shade=False,
                                                render=False,
                                                reset_camera=False)

            with self._plotter.allowNonblockingCalls():
                setActorUserTransform(self._volActor, self.session.MRI.dataToScannerTransf)
                self.plotter.reset_camera()
                # self.plotter.camera.zoom('tight')

        logger.debug('Setting crosshairs for {} plot'.format(self.label))
        lineLength = 300  # TODO: scale by image size
        crosshairAxes = 'xyz'
        centerGapLength = 0
        for axis in crosshairAxes:
            mask = np.zeros((1, 3))
            mask[0, 'xyz'.index(axis)] = 1
            for iDir, dir in enumerate((-1, 1)):
                pts = dir*np.asarray([centerGapLength/2, lineLength])[:, np.newaxis] * mask + self._sliceOrigin
                lineKey = 'Crosshair_{}_{}_{}'.format(self.label, axis, iDir)
                if not self._plotterInitialized:
                    line = self._plotter.add_lines(pts, color='#11DD11', width=2, name=lineKey)
                    with self._plotter.allowNonblockingCalls():
                        line.SetUseBounds(False)  # don't include for determining camera zoom, etc.
                    self._lineActors[lineKey] = line
                else:
                    with self._plotter.allowNonblockingCalls():
                        logger.debug('Moving previous crosshairs')
                        line = self._lineActors[lineKey]
                        pts_pv = pv.lines_from_points(pts)
                        line.GetMapper().SetInputData(pts_pv)

        self._plotterInitialized = True

        with self.plotter.allowNonblockingCalls():
            self.plotter.render()

    def _onMRIClimChanged(self, dim: str):
        if dim == '3D':
            self._queueRedraw(which='updateClim')

    def _updateClim(self):
        if not self.finishedAsyncInit.is_set():
            # plotter not available yet
            return

        if not self._plotterInitialized:
            # initial plot not created yet
            return

        if self.session is not None and self.session.MRI.isSet:
            with self.plotter.allowNonblockingCalls():
                self._plotter.updateScalarBarRangeWithVol(
                    clim=self.session.MRI.clim3D,
                    volumeKey='MRI',
                    volume=self.session.MRI.dataAsUniformGrid.gaussian_smooth()
                )
                self._plotter.render()



import sys

from qtpy.QtCore import Qt
from qtpy.QtGui import QDropEvent
from qtpy.QtWidgets import QTableWidget, QAbstractItemView, QTableWidgetItem

from NaviNIBS.util.Signaler import Signal


class QTableWidgetDragRows(QTableWidget):
    """
    Adapted from https://stackoverflow.com/a/43789304
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.viewport().setAcceptDrops(True)
        self.setDragDropOverwriteMode(False)
        self.setDropIndicatorShown(True)

        self.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setDragDropMode(QAbstractItemView.InternalMove)

        self.dropInProgress = False
        self.sigDragAndDropReordered = Signal()

    def dropEvent(self, event: QDropEvent):
        if not event.isAccepted() and event.source() == self:
            self.dropInProgress = True
            drop_row = self.drop_on(event)

            rows = sorted(set(item.row() for item in self.selectedItems()))
            rows_to_move = [[QTableWidgetItem(self.item(row_index, column_index)) for column_index in range(self.columnCount())]
                            for row_index in rows]
            for row_index in reversed(rows):
                self.removeRow(row_index)
                if row_index < drop_row:
                    drop_row -= 1

            for row_index, data in enumerate(rows_to_move):
                row_index += drop_row
                self.insertRow(row_index)
                for column_index, column_data in enumerate(data):
                    self.setItem(row_index, column_index, column_data)
            event.accept()
            for row_index in range(len(rows_to_move)):
                self.item(drop_row + row_index, 0).setSelected(True)
                self.item(drop_row + row_index, 1).setSelected(True)

            self.dropInProgress = False
            self.sigDragAndDropReordered.emit()

        super().dropEvent(event)

    def drop_on(self, event):
        index = self.indexAt(event.pos())
        if not index.isValid():
            return self.rowCount()

        return index.row() + 1 if self.is_below(event.pos(), index) else index.row()

    def is_below(self, pos, index):
        rect = self.visualRect(index)
        margin = 2
        if pos.y() - rect.top() < margin:
            return False
        elif rect.bottom() - pos.y() < margin:
            return True
        # noinspection PyTypeChecker
        return rect.contains(pos, True) and not (int(self.model().flags(index)) & Qt.ItemIsDropEnabled) and pos.y() >= rect.center().y()



from __future__ import annotations

import asyncio

import attrs
import logging
import numpy as np
import pyvista as pv
import pyvistaqt as pvqt
import qtawesome as qta
from qtpy import QtWidgets, QtGui, QtCore
import typing as tp
from typing import ClassVar

from NaviNIBS.Navigator.GUI.ViewPanels.NavigatePanel.NavigationView import TargetingCoordinator
from NaviNIBS.util import classproperty

from NaviNIBS.util.GUI.QueuedRedrawMixin import QueuedRedrawMixin
from NaviNIBS.util.pyvista import Actor, setActorUserTransform, concatenateLineSegments
from NaviNIBS.util.pyvista import DefaultBackgroundPlotter, RemotePlotterProxy



logger = logging.getLogger(__name__)


Transform = np.ndarray


@attrs.define
class LegendEntry:
    _label: str
    _color: tp.Any

    """
    Note: attributes below are shared by entire legend, so each entry just provides its preference but may not 
    actually be granted these values. Or in the future, may implement multiple / more flexible pyvista legends
    to support showing these on a per-entry basis
    """
    _bcolor: tp.Any | None = None
    _border: bool = False
    _loc: str = 'upper right'
    _face: str | pv.PolyData | None = 'triangle'

    _legendActor: pv.Actor | None = attrs.field(init=False, default=None)

    def __attrs_post_init__(self):
        pass

    @property
    def label(self):
        return self._label

    @property
    def color(self):
        return self._color

    @property
    def bcolor(self):
        return

    @property
    def border(self):
        return self._border

    @property
    def loc(self):
        return self._loc

    @property
    def face(self):
        return self._face


@attrs.define
class ViewLayer:
    _key: str
    _type: ClassVar[str]
    _coordinator: TargetingCoordinator

    def __attrs_post_init__(self):
        pass

    @classproperty
    def type(cls):
        return cls._type


@attrs.define
class PlotViewLayer(ViewLayer, QueuedRedrawMixin):
    _plotter: DefaultBackgroundPlotter  # note that this one plotter may be shared between multiple ViewLayers
    _plotInSpace: str = 'MRI'

    _actors: dict[str, Actor | None] = attrs.field(init=False, factory=dict)
    _legendEntries: list[LegendEntry] = attrs.field(init=False, factory=dict)

    def __attrs_post_init__(self):
        ViewLayer.__attrs_post_init__(self)
        QueuedRedrawMixin.__attrs_post_init__(self)
        self._redraw('all')

    @property
    def legendEntries(self):
        return self._legendEntries

    def _registerLegendEntry(self, entry: LegendEntry):
        self._legendEntries.append(entry)

    def _redraw(self, which: tp.Union[tp.Optional[str], tp.List[str, ...]] = None):
        QueuedRedrawMixin._redraw(self, which=which)

        if isinstance(self._plotter, RemotePlotterProxy) and not self._plotter.isReadyEvent.is_set():
            # remote plotter not ready yet
            return

        #logger.debug('redraw {}'.format(which))

        if which is None:
            which = 'all'
            self._redraw(which=which)
            return

        if not isinstance(which, str):
            for subWhich in which:
                self._redraw(which=subWhich)
            return

        # subclass should handle the rest

    def _getActorKey(self, subKey: str) -> str:
        return self._key + '_' + subKey  # make actor keys unique across multiple layers in the same plotter



import numpy as np
import math


def getNiceTicksForRange(minVal: float, maxVal: float, approxNumTicks: int = 5) -> np.ndarray:
    return _getNiceTicksForRangeV2(minVal, maxVal, approxNumTicks)


def _getNiceTicksForRangeV1(minVal: float, maxVal: float, approxNumTicks: int = 5) -> np.ndarray:
    """
    Get nice ticks for a range of values. 
    """
    boundsWidth = maxVal - minVal
    numTicks = approxNumTicks
    tickOrderMagnitude = np.floor(np.log10(boundsWidth / numTicks))
    tickMultiple = 10 ** tickOrderMagnitude
    ticksMin = np.floor(minVal / tickMultiple) * tickMultiple
    ticksMax = np.ceil(maxVal / tickMultiple) * tickMultiple

    numTickMultiples = round((ticksMin - ticksMax) / tickMultiple)
    if numTickMultiples % numTicks != 0:
        # if ticks are not on nice intervals, adjust bounds to make them so

        numToAdd = numTicks - (numTickMultiples % numTicks)

        penalty_iFAdd = 0
        ticksMin_ifAdd = ticksMin
        ticksMax_ifAdd = ticksMax
        for iA in range(numToAdd):
            penalty_low = abs(ticksMin_ifAdd - tickMultiple - minVal)
            penalty_high = abs(maxVal - ticksMax_ifAdd - tickMultiple)
            if penalty_low < penalty_high:
                ticksMin_ifAdd -= tickMultiple
                penalty_iFAdd += penalty_low
            else:
                ticksMax_ifAdd += tickMultiple
                penalty_iFAdd += penalty_high

        numToRemove = numTickMultiples % numTicks

        penalty_iFRemove = 0
        ticksMin_ifRemove = ticksMin
        ticksMax_ifRemove = ticksMax
        for iR in range(numToRemove):
            penalty_low = abs(minVal - ticksMin_ifRemove + tickMultiple)
            penalty_high = abs(ticksMax_ifRemove + tickMultiple - maxVal)
            if penalty_low < penalty_high:
                ticksMin_ifRemove += tickMultiple
                penalty_iFRemove += penalty_low
            else:
                ticksMax_ifRemove -= tickMultiple
                penalty_iFRemove += penalty_high

        if penalty_iFAdd < 5 * penalty_iFRemove:
            ticksMin = ticksMin_ifAdd
            ticksMax = ticksMax_ifAdd
        else:
            ticksMin = ticksMin_ifRemove
            ticksMax = ticksMax_ifRemove

    return np.linspace(ticksMin, ticksMax, numTicks)


def _getNiceTicksForRangeV2(minVal: float, maxVal: float, approxNumTicks: int | None = None) -> np.ndarray:
    """
    Adapted from https://stackoverflow.com/a/73313693/2388228
    """
    retpoints = []
    data_range = maxVal - minVal
    lower_bound = minVal - data_range / 8
    upper_bound = maxVal + data_range / 8
    view_range = upper_bound - lower_bound
    num = lower_bound
    n = math.floor(math.log10(view_range) - 1)
    interval = 10 ** n
    num_ticks = 1
    while num <= upper_bound:
        num += interval
        num_ticks += 1
        if num_ticks > approxNumTicks * 1.5:
            if interval == 10 ** n:
                interval = 2 * 10 ** n
            elif interval == 2 * 10 ** n:
                interval = 4 * 10 ** n
            elif interval == 4 * 10 ** n:
                interval = 5 * 10 ** n
            else:
                n += 1
                interval = 10 ** n
            num = lower_bound
            num_ticks = 1
    if view_range >= 10:
        copy_interval = interval
    else:
        if interval == 10 ** n:
            copy_interval = 1
        elif interval == 2 * 10 ** n:
            copy_interval = 2
        elif interval == 4 * 10 ** n:
            copy_interval = 4
        else:
            copy_interval = 5
    first_val = 0
    prev_val = 0
    times = 0
    temp_log = math.log10(interval)
    if math.isclose(lower_bound, 0):
        first_val = 0
    elif lower_bound < 0:
        if upper_bound < -2 * interval:
            if n < 0:
                copy_ub = round(upper_bound * 10 ** (abs(temp_log) + 1))
                times = copy_ub // round(interval * 10 ** (abs(temp_log) + 1)) + 2
            else:
                times = upper_bound // round(interval) + 2
        while first_val >= lower_bound:
            prev_val = first_val
            first_val = times * copy_interval
            if n < 0:
                first_val *= (10 ** n)
            times -= 1
        first_val = prev_val
        times += 3
    else:
        if lower_bound > 2 * interval:
            if n < 0:
                copy_ub = round(lower_bound * 10 ** (abs(temp_log) + 1))
                times = copy_ub // round(interval * 10 ** (abs(temp_log) + 1)) - 2
            else:
                times = lower_bound // round(interval) - 2
        while first_val < lower_bound:
            first_val = times * copy_interval
            if n < 0:
                first_val *= (10 ** n)
            times += 1
    if n < 0:
        retpoints.append(first_val)
    else:
        retpoints.append(round(first_val))
    val = first_val
    times = 1
    while val <= upper_bound:
        val = first_val + times * interval
        if n < 0:
            retpoints.append(val)
        else:
            retpoints.append(round(val))
        times += 1
    retpoints.pop()
    return np.asarray(retpoints)


from __future__ import annotations

import attrs
import logging
import numpy as np
import pyvista as pv
import typing as tp
from typing import ClassVar

from . import PlotViewLayer
from NaviNIBS.Navigator.TargetingCoordinator import ProjectionSpecification

logger = logging.getLogger(__name__)


Transform = np.ndarray


@attrs.define(kw_only=True)
class TargetingErrorLineLayer(PlotViewLayer):
    _type: ClassVar[str] = 'TargetingErrorLine'

    _targetDepth: tp.Union[str, ProjectionSpecification] = 'target'
    _coilDepth: tp.Union[str, ProjectionSpecification] = 'target'

    _color: str = '#ff0000'
    _opacity: float = 0.5
    _lineWidth: float = 4.

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self._coordinator.sigCurrentTargetChanged.connect(lambda: self._queueRedraw(which='updatePosition'))
        self._coordinator.sigCurrentCoilPositionChanged.connect(lambda: self._queueRedraw(which=['updatePosition']))

    def _redraw(self, which: tp.Union[tp.Optional[str], tp.List[str, ...]] = None):
        super()._redraw(which=which)

        if not isinstance(which, str):
            # assume parent call above triggered appropriate redraws
            return

        if which == 'all':
            which = ['initLine']
            self._redraw(which=which)
            return

        actorKey = self._getActorKey('line')

        if which == 'initLine':
            # start with dummy position, to be updated later
            pts_line = np.asarray([[0., 0., 0.], [1., 1., 1.]])

            self._actors[actorKey] = self._plotter.add_lines(pts_line,
                                                             color=self._color,
                                                             width=self._lineWidth,
                                                             name=actorKey)

            self._redraw(which='updatePosition')

        elif which == 'updatePosition':
            actor = self._actors[actorKey]

            doShow = True
            pt_coil = self._coordinator.getTargetingCoord(orientation='coil', depth=self._coilDepth)
            if pt_coil is None:
                doShow = False
            else:
                pt_target = self._coordinator.getTargetingCoord(orientation='target',  depth=self._targetDepth)
                if pt_target is None:
                    doShow = False

            if not doShow:
                if actor.GetVisibility():
                    with self._plotter.allowNonblockingCalls():
                        actor.VisibilityOff()
                return

            pts_pv = pv.lines_from_points(np.vstack([pt_coil, pt_target]))

            with self._plotter.allowNonblockingCalls():
                actor.GetMapper().SetInputData(pts_pv)

                if not actor.GetVisibility():
                    actor.VisibilityOn()

                self._plotter.render()
        else:
            raise NotImplementedError(f'Unexpected redraw which: {which}')




import attrs

from NaviNIBS.Navigator.GUI.CollectionModels import CollectionTableModel, K, logger
from NaviNIBS.Navigator.Model.Tools import Tool, Tools


@attrs.define(slots=False)
class ToolsTableModel(CollectionTableModel[int, Tools, Tool]):
    _collection: Tools = attrs.field(init=False, repr=False)

    def __attrs_post_init__(self):
        self._collection = self._session.tools

        self._columns = [
            'label',
            'isActive',
            'doRenderTool',
            'doRenderTracker',
            'doShowTrackingState'
        ]

        self._attrColumns = self._columns.copy()

        self._columnLabels = dict(
            label='Label',
            isActive='Active',
            doRenderTool='Render tool',
            doRenderTracker='Render tracker',
            doShowTrackingState='Show tracking state'
        )

        self._boolColumns = [
            'isActive',
            'doRenderTool',
            'doRenderTracker',
            'doShowTrackingState'
        ]

        self._editableColumns = self._boolColumns.copy()

        self._collection.sigItemsAboutToChange.connect(self._onCollectionAboutToChange)
        self._collection.sigItemsChanged.connect(self._onCollectionChanged)

        super().__attrs_post_init__()








import attr
import asyncio
import json
import logging
from math import ceil
import socket
import typing as tp
import unittest
import zmq
import zmq.asyncio as azmq

from . import exceptionToStr
from . import ZMQAsyncioFix

from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError

logger = logging.getLogger(__name__)


class InvalidMessageError(Exception):
    pass


def _msgSizeInGB(msg: tp.List[bytes]) -> float:
    return sum(len(item) for item in msg) / 1.e9


@attr.s(auto_attribs=True)
class ZMQConnectorServer:
    _obj: tp.Any
    _reqRepPort: int
    _pubSubPort: tp.Optional[int] = None  # if not specified, will not be able to publish
    _bindAddr: str = '127.0.0.1'  # set to '*' to bind all interfaces
    _ctx: tp.Optional[azmq.Context] = attr.ib(factory=azmq.Context)

    _onOtherRequestReceived: tp.Optional[tp.Callable[[tp.List[bytes]], tp.List[bytes]]] = None

    _repSocket: tp.Optional[azmq.Socket] = attr.ib(init=False, default=None)
    _pubSocket: tp.Optional[azmq.Socket] = attr.ib(init=False, default=None)
    _asyncPollingTask: asyncio.Task = attr.ib(init=False)

    def __attrs_post_init__(self):
        logger.debug('Binding rep socket on port %d' % (self._reqRepPort,))
        self._repSocket = self._ctx.socket(zmq.REP)
        self._repSocket.bind('tcp://%s:%d' % (self._bindAddr, self._reqRepPort))

        if self._pubSubPort is not None:
            logger.debug('Binding pub socket on port %d' % (self._pubSubPort,))
            self._pubSocket = self._ctx.socket(zmq.PUB)
            self._pubSocket.linger = 0
            self._pubSocket.bind('tcp://%s:%d' % (self._bindAddr, self._pubSubPort))

        self._asyncPollingTask = asyncio.create_task(asyncTryAndLogExceptionOnError(self._asyncPoll))

    def __del__(self):
        logger.debug("Deleting ZMQConnectorServer")
        self.close()

    def close(self):
        if self._repSocket is not None:
            self._repSocket.linger = 0
            self._repSocket.close(0)
            self._repSocket = None

        if self._pubSocket is not None:
            self._asyncPollingTask.cancel()
            self._pubSocket.linger = 0
            self._pubSocket.close(0)
            self._pubSocket = None

        if self._ctx is not None:
            self._ctx.term()
            self._ctx = None

    async def _processRequest(self, req: tp.List[bytes]) -> tp.List[bytes]:

        msgType = req[0]

        def encodeResponse(resp: tp.Any) -> bytes:
            try:
                return [json.dumps(resp).encode('utf-8')]
            except TypeError as e:
                logger.error(f'Problem serializing request response: {resp}')
                raise e

        logger.debug("Processing request of type %s" % msgType)
        try:
            if msgType == b'ping':
                if len(req) == 1:
                    logger.debug('Responding to ping with pong')
                    return [b'pong']
                else:
                    raise InvalidMessageError()

            elif msgType in (b'get', b'set', b'call', b'asyncCall'):
                assert len(req) == 2
                req = json.loads(req[1].decode('utf-8'))

                if not isinstance(req, list):
                    raise InvalidMessageError()

                if msgType == b'get':
                    whatToGet = req[0]
                    if not isinstance(whatToGet, list):
                        raise InvalidMessageError()
                    resp = []
                    for what in whatToGet:
                        val = getattr(self._obj, what)
                        logger.debug("Got %s = %s" % (what, val))
                        resp.append(val)
                    return encodeResponse(resp)

                elif msgType == b'set':
                    whatToSet = req[0]
                    if not isinstance(whatToSet,dict):
                        raise InvalidMessageError()
                    for what, val in whatToSet.items():
                        if not hasattr(self._obj, what):
                            logger.warning("ZMQConnector setting new attribute via request: %s = %s" % (what, val))
                        try:
                            setattr(self._obj, what, val)
                        except Exception as e:
                            logger.error('Failed to set %s to %s: %s' % (what, val, exceptionToStr(e)))
                            return encodeResponse(['__Error__', 'SetFailedError', exceptionToStr(e)])

                        logger.debug("Set %s = %s" % (what, val))
                    return encodeResponse('success')

                elif msgType in (b'call', b'asyncCall'):
                    whatToCall = req[0]
                    if not isinstance(whatToCall, dict):
                        raise InvalidMessageError()

                    try:
                        method = whatToCall['method']
                    except KeyError as e:
                        raise InvalidMessageError()

                    args = whatToCall.get('args', list())
                    kwargs = whatToCall.get('kwargs', dict())

                    if not isinstance(args, list):
                        raise InvalidMessageError()

                    if not isinstance(kwargs, dict):
                        raise InvalidMessageError()

                    toCall = getattr(self._obj, method)

                    if msgType == b'call':
                        ret = toCall(*args, **kwargs)
                    else:
                        ret = await toCall(*args, **kwargs)

                    return encodeResponse(['success', ret])
                else:
                    raise NotImplementedError()

            elif msgType in (b'getB', b'bSet'):
                if msgType == b'getB':
                    assert len(req) == 2
                    req = json.loads(req[1].decode('utf-8'))

                    if not isinstance(req, list):
                        raise InvalidMessageError()
                    whatToGet = req[0]
                    if not isinstance(whatToGet, list):
                        raise InvalidMessageError()
                    resp = []
                    assert len(whatToGet) == 1
                    val = getattr(self._obj, whatToGet[0])
                    assert isinstance(val, list)
                    for subval in val:
                        assert isinstance(subval, bytes)
                    logger.debug("Got %s = <list of bytes, size = %.3f GB>" % (whatToGet[0], _msgSizeInGB(val)))
                    return val
                elif msgType == b'bSet':
                    assert len(req) > 2
                    what = req[1].decode('utf-8')
                    val = req[2:]
                    try:
                        setattr(self._obj, what, val)
                    except Exception as e:
                        logger.error('Failed to set %s: %s' % (what, exceptionToStr(e)))
                        return encodeResponse(['__Error__', 'SetFailedError', exceptionToStr(e)])
                    logger.debug('Set %s = <list of bytes, size = %.3f GB>' % (what, _msgSizeInGB(val)))
                    return encodeResponse('success')
                else:
                    raise NotImplementedError()

            else:
                if self._onOtherRequestReceived is not None:
                    # allow custom message handlers
                    return self._onOtherRequestReceived(req)
                else:
                    raise InvalidMessageError()

        except InvalidMessageError as e:
            return encodeResponse(['__Error__', 'InvalidMessageError', exceptionToStr(e)])
        except AttributeError as e:
            return encodeResponse(['__Error__', 'AttributeError', exceptionToStr(e)])

    def publish(self, msg: tp.List[bytes]):
        assert self._pubSocket is not None
        self._pubSocket.send_multipart(msg)

    async def _asyncPoll(self):
        poller = azmq.Poller()
        poller.register(self._repSocket, zmq.POLLIN)
        while True:
            socks = dict(await poller.poll())
            if self._repSocket in socks:
                req = await self._repSocket.recv_multipart()
                try:
                    #logger.debug('Processing request')
                    resp = await self._processRequest(req)
                    #logger.debug('Done processing request')

                    await self._repSocket.send_multipart(resp)
                except Exception as e:
                    logger.error("Unhandled exception in ZMQConnectorServer:\n %s" % (exceptionToStr(e),))
                    await self._repSocket.send_multipart([json.dumps(
                        ['__Error__', 'CriticalFailure', exceptionToStr(e)]).encode('utf-8')])
                    #raise e


class RemoteError(Exception):
    pass


@attr.s(auto_attribs=True, cmp=False)
class ZMQConnectorClient:
    _reqRepPort: int
    _pubSubPort: tp.Optional[int] = None  # if not specified, will not be able to subscribe
    _connAddr: str = '127.0.0.1'
    _ctx: tp.Optional[zmq.Context] = attr.ib(factory=zmq.Context)
    _actx: tp.Optional[azmq.Context] = attr.ib(factory=azmq.Context)
    onMessagePublished: tp.Optional[tp.Callable[[tp.List[bytes]], None]] = None

    _reqSocket: zmq.Socket = attr.ib(init=False)
    _areqSocket: azmq.Socket = attr.ib(init=False)
    _areqLock: asyncio.Lock | None = attr.ib(init=False, default=None)
    _subSocket: tp.Optional[azmq.Socket] = attr.ib(init=False, default=None)

    _allowAsyncCalls: bool = False
    _allowSyncCalls: bool = True

    def __attrs_post_init__(self):
        self._connect()

    def __del__(self):
        logger.debug('Deleting ZMQConnectorClient')
        self.close()

    @property
    def connAddr(self):
        return self._connAddr

    def _connect(self):
        if self._allowSyncCalls:
            logger.debug('Initializing req socket')
            self._reqSocket = self._ctx.socket(zmq.REQ)
            logger.debug('Connecting req socket')
            self._reqSocket.connect('tcp://%s:%d' % (self._connAddr, self._reqRepPort))
        if self._allowAsyncCalls:
            logger.debug('Connecting async req socket')
            self._areqSocket = self._actx.socket(zmq.REQ)
            self._areqSocket.connect('tcp://%s:%d' % (self._connAddr, self._reqRepPort))
            if self._areqLock is None:
                self._areqLock = asyncio.Lock()

        if self._pubSubPort is not None:
            logger.debug('Setting up subscribe')
            logger.debug('Connecting sub socket')
            self._subSocket = self._actx.socket(zmq.SUB)
            self._subSocket.connect('tcp://%s:%d' % (self._connAddr, self._pubSubPort))
            self._subSocket.setsockopt(zmq.SUBSCRIBE, b'')

            asyncio.create_task(asyncTryAndLogExceptionOnError(self._asyncPoll))  # only need to poll if checking for published updates
            logger.debug('Finished setting up subscribe')

    def close(self, linger=0):
        if self._allowSyncCalls and self._reqSocket is not None:
            self._reqSocket.close(linger)
            self._reqSocket = None
        if self._allowAsyncCalls and self._areqSocket is not None:
            self._areqSocket.close(linger)
            self._areqSocket = None
        if self._subSocket is not None:
            self._subSocket.close(linger)
            self._subSocket = None
        if self._ctx is not None:
            self._ctx.term()
            self._ctx = None
        if self._actx is not None:
            self._actx.term()
            self._actx = None

    async def _asyncPoll(self):
        poller = azmq.Poller()
        poller.register(self._subSocket, zmq.POLLIN)

        while True:
            #logger.debug('Polling for updates')
            socks = dict(await poller.poll())
            if self._subSocket in socks:
                rawMsg = await self._subSocket.recv_multipart()
                assert self.onMessagePublished is not None
                self.onMessagePublished(rawMsg)

    async def send_async(self, msg: tp.List[bytes]) -> tp.List[bytes]:
        assert self._allowAsyncCalls

        logger.debug('About to wait for areqLock')
        async with self._areqLock:
            logger.debug('Acquired areqLock')
            await self._areqSocket.send_multipart(msg)
            logger.debug('Waiting for areq response')
            resp = await self._areqSocket.recv_multipart()
            logger.debug('Received response')

        return resp

    def send(self, msg: tp.List[bytes]) -> tp.List[bytes]:
        # assume this message doesn't conflict with format of built-in messages and just send directly
        logger.debug('Sending msg')
        self._reqSocket.send_multipart(msg)
        resp = self._reqSocket.recv_multipart()
        logger.debug('Received response')
        return resp

    async def _sendTypePlusBody_receive_async(self, msgType: str, obj: tp.Any,
                                              doEncode: bool = True, doDecodeResponse: bool = True) -> tp.Any:
        # (normally could have been handled by socket.send_json and socket.recv_json
        #  but with some protocols using send_multipart and recv_multipart this
        #  is acting as wrapper)
        if doEncode:
            resp = await self.send_async([msgType.encode('utf-8'),
                                          json.dumps(obj).encode('utf-8')])
        else:
            assert isinstance(obj, list)
            for part in obj:
                assert isinstance(part, bytes)

            resp = await self.send_async([msgType.encode('utf-8')] + obj)

        if doDecodeResponse:
            if len(resp) != 1:
                raise InvalidMessageError('Unexpected message format received from remote')

            resp = json.loads(resp[0].decode('utf-8'))

        return resp

    def _sendTypePlusBody_receive(self, msgType: str, obj: tp.Any,
                                  doEncode: bool = True, doDecodeResponse: bool = True) -> tp.Any:
        # (normally could have been handled by socket.send_json and socket.recv_json
        #  but with some protocols using send_multipart and recv_multipart this
        #  is acting as wrapper)
        if doEncode:
            resp = self.send([msgType.encode('utf-8'),
                                            json.dumps(obj).encode('utf-8')])
        else:
            assert isinstance(obj, list)
            for part in obj:
                assert isinstance(part, bytes)

            resp = self.send([msgType.encode('utf-8')] + obj)

        if doDecodeResponse:
            if len(resp) != 1:
                raise InvalidMessageError('Unexpected message format received from remote')

            resp = json.loads(resp[0].decode('utf-8'))

        return resp

    async def get_async(self, item: str, raw: bool = False) -> tp.Any:
        assert isinstance(item, str)

        logger.debug('Sending get request for %s' % item)

        resp = await self._sendTypePlusBody_receive_async('get' if not raw else 'getB', [[item]], doDecodeResponse=not raw)

        logger.debug('Received response')

        if not raw:
            if isinstance(resp, list):
                if resp[0] == '__Error__':
                    if resp[1] == 'AttributeError':
                        raise AttributeError("Server could not get attribute '%s'" % item)
                    else:
                        raise RemoteError("Unhandled exception '%s' trying to get attribute '%s' from server.\n%s" % (
                            resp[1], item, resp[2]))

            if len(resp) != 1:
                logger.debug('problem')  # TODO: debug, delete
                raise InvalidMessageError()

            resp = resp[0]

            logger.debug("Got %s = %s" % (item, resp))
        else:
            logger.debug("Got %s = <list of bytes, size = %.2f GB>" % (item, _msgSizeInGB(resp)))

        return resp

    def get(self, item: str, raw: bool = False) -> tp.Any:
        assert isinstance(item, str)

        logger.debug('Sending get request for %s' % item)

        resp = self._sendTypePlusBody_receive('get' if not raw else 'getB', [[item]], doDecodeResponse=not raw)

        if not raw:
            if isinstance(resp, list):
                if resp[0] == '__Error__':
                    if resp[1] == 'AttributeError':
                        raise AttributeError("Server could not get attribute '%s'" % item)
                    else:
                        raise RemoteError("Unhandled exception '%s' trying to get attribute '%s' from server.\n%s" % (
                            resp[1], item, resp[2]))

            if len(resp) != 1:
                logger.debug('problem')  # TODO: debug, delete
                raise InvalidMessageError()

            resp = resp[0]

            logger.debug("Got %s = %s" % (item, resp))
        else:
            logger.debug('Got %s = <list of bytes with %d items>' % (item, len(resp)))

        return resp

    async def set_async(self, key, value, raw: bool = False):
        if not raw:
            resp = await self._sendTypePlusBody_receive_async('set', [{key: value}])
        else:
            assert isinstance(value, list)
            for subvalue in value:
                assert isinstance(subvalue, bytes)
            resp = await self._sendTypePlusBody_receive_async('bSet', [key.encode('utf-8')] + value, doEncode=False)

        if isinstance(resp, list):
            if resp[0] == '__Error__':
                raise RemoteError("Unhandled exception '%s' trying to set attribute '%s' on server" % (resp[1], key))

        if not isinstance(resp, str) or resp != 'success':
            raise RemoteError("Failed to set '%s' on server" % key)

        if not raw:
            logger.debug("Set %s = %s" % (key, value))
        else:
            logger.debug('Set %s = <list of bytes, size = %.3f GB' % (key, _msgSizeInGB(value)))

    def set(self, key, value, raw: bool = False):
        if not raw:
            resp = self._sendTypePlusBody_receive('set', [{key: value}])
        else:
            assert isinstance(value, list)
            for subvalue in value:
                assert isinstance(subvalue, bytes)
            resp = self._sendTypePlusBody_receive('bSet', [key.encode('utf-8')] + value, doEncode=False)

        if isinstance(resp, list):
            if resp[0] == '__Error__':
                raise RemoteError("Unhandled exception '%s' trying to set attribute '%s' on server:\n %s" % (resp[1], key, resp[2]))

        if not isinstance(resp, str) or resp != 'success':
            raise RemoteError("Failed to set '%s' on server" % key)

        if not raw:
            logger.debug("Set %s = %s" % (key, value))
        else:
            logger.debug('Set %s = <list of bytes, size = %.3f GB' % (key, _msgSizeInGB(value)))

    async def _call_async(self, method: str, doAsync: bool, *args, **kwargs):
        resp = await self._sendTypePlusBody_receive_async('asyncCall' if doAsync else 'call', [dict(
            method=method,
            args=args,
            kwargs=kwargs
        )])

        assert isinstance(resp, list)

        if resp[0] == '__Error__':
            raise RemoteError(
                "Unhandled exception '%s' trying to call method '%s' on server:\n%s" % (resp[1], method, resp[2]))

        if resp[0] != 'success':
            raise RemoteError("Failed to call '%s' on server" % method)

        return resp[1]

    def _call(self, method: str, doAsync: bool, *args, **kwargs):
        resp = self._sendTypePlusBody_receive('asyncCall' if doAsync else 'call', [dict(
            method=method,
            args=args,
            kwargs=kwargs
        )])

        assert isinstance(resp, list)

        if resp[0] == '__Error__':
            raise RemoteError(
                "Unhandled exception '%s' trying to call method '%s' on server:\n%s" % (resp[1], method, resp[2]))

        if resp[0] != 'success':
            raise RemoteError("Failed to call '%s' on server" % method)

        return resp[1]

    async def call_async(self, method: str, *args, **kwargs):
        """
        Asynchronously on the client, make a synchronous call on the server
        """
        return await self._call_async(method, False, *args, **kwargs)

    def call(self, method: str, *args, **kwargs):
        return self._call(method, False, *args, **kwargs)

    async def callAsync_async(self, method: str, *args, **kwargs):
        """"
        Asynchronously on the client, make an asynchronous call on the server
        """
        return await self._call_async(method, True, *args, **kwargs)

    def callAsync(self, method: str, *args, **kwargs):
        """
        Synchronously on the client, make an asynchronous call on the server
        """
        return self._call(method, True, *args, **kwargs)

    async def ping_async(self, timeout=100, numTries=2):
        """
        :param timeout: time in ms
        :param numTries:
        :return: None

        Raises TimeoutError if ping fails
        """
        for i in range(numTries):
            async with self._areqLock:
                await self._areqSocket.send_multipart([b'ping'])
                logger.debug("Sent ping")

                poll = azmq.Poller()
                poll.register(self._areqSocket, zmq.POLLIN)
                socks = dict(await poll.poll(timeout / numTries))
                if socks.get(self._areqSocket) == zmq.POLLIN:
                    resp = await self._areqSocket.recv_multipart()
                    assert len(resp) == 1 and resp[0] == b'pong'
                    logger.debug("Received pong")
                    return
                else:
                    logger.debug('Ping timed out, closing connection')
                    self._areqSocket.setsockopt(zmq.LINGER, 0)
                    self._areqSocket.close()
                    poll.unregister(self._areqSocket)

                    if self._allowSyncCalls:
                        self._reqSocket.setsockopt(zmq.LINGER, 0)
                        self._reqSocket.close()

                    if self._subSocket is not None:
                        self._subSocket.setsockopt(zmq.LINGER, 0)
                        self._subSocket.close()

                    self._connect()

        raise TimeoutError()

    def ping(self, timeout=100, numTries=2):
        for i in range(numTries):
            self._reqSocket.send_multipart([b'ping'])
            logger.debug("Sent ping")

            poll = zmq.Poller()
            poll.register(self._reqSocket, zmq.POLLIN)
            socks = dict(poll.poll(timeout / numTries))
            if socks.get(self._reqSocket) == zmq.POLLIN:
                resp = self._reqSocket.recv_multipart()
                assert len(resp) == 1 and resp[0] == b'pong'
                logger.debug("Received pong")
                return
            else:
                logger.debug('Ping timed out, closing connection')
                self._reqSocket.setsockopt(zmq.LINGER, 0)
                self._reqSocket.close()
                poll.unregister(self._reqSocket)

                if self._allowAsyncCalls:
                    self._areqSocket.setsockopt(zmq.LINGER, 0)
                    self._areqSocket.close()

                if self._subSocket is not None:
                    self._subSocket.setsockopt(zmq.LINGER, 0)
                    self._subSocket.close()

                self._connect()

        raise TimeoutError()


def checkIfPortAvailable(port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        logger.debug('Created test socket on port %d' % (port,))  # TODO: debug, delete
        s.settimeout(0.01)
        return s.connect_ex(('localhost', port)) != 0


usedPorts = set()


def getNewPort(minAssignedPort: int = 5000) -> int:
    global usedPorts
    port = minAssignedPort

    # Note: this only checks that localhost can bind to the port. If wanting to get a port
    #  for a remote machine (i.e. not localhost) to bind to, this does not ensure availability.
    logger.debug('Checking to make sure port is available')
    while True:
        if port in usedPorts or not checkIfPortAvailable(port):
            # mark port as in use so we don't keep trying it in the future
            logger.debug('Port %d is already in use. Skipping.' % (port,))
            usedPorts.add(port)
            port += 1
            continue
        else:
            # found available port
            logger.debug('Found available port: %d' % (port,))
            # assume it will be used
            usedPorts.add(port)
            return port


class Test_ZMQConnector(unittest.TestCase):
    class state:
        a: int
        b: str
        c: tp.List[bytes]

        async def respondAfterDelay(self, delay: float = 1., toReturn=True):
            await asyncio.sleep(delay)
            return toReturn

    def _test_basic_sync(self, client):
        clientState = Test_ZMQConnector.state()
        clientState.a = client.get('a')
        clientState.b = client.get('b')
        clientState.c = 3
        client.set('c', clientState.c)
        return clientState

    async def _test_basic(self, doAsync):

        reqRepPort = 7897

        serverState = Test_ZMQConnector.state()
        serverState.a = 1
        serverState.b = 'two'
        addrKey = 'testZMQConnector'
        server = ZMQConnectorServer(obj=serverState, reqRepPort=reqRepPort)
        await asyncio.sleep(1)
        logger.debug("Finished initializing server")
        client = ZMQConnectorClient(reqRepPort=reqRepPort, allowAsyncCalls=doAsync)

        await asyncio.sleep(1)

        logger.debug("Getting client state")
        if doAsync:
            clientState = Test_ZMQConnector.state()
            clientState.a = await client.get_async('a')
            clientState.b = await client.get_async('b')
            clientState.c = 3
            await client.set_async('c', clientState.c)
        else:
            # for testing, wrap blocking non-async call in async call here so that server keeps running at the same time
            # (in practice, the server would already be running in a separate thread)
            loop = asyncio.get_running_loop()
            clientState = await loop.run_in_executor(None, lambda: self._test_basic_sync(client))

        await asyncio.sleep(1)

        self.assertEqual(serverState.a, clientState.a)
        self.assertEqual(serverState.b, clientState.b)
        self.assertEqual(serverState.c, clientState.c)

        server.close() # needed for successive tests
        del server
        await asyncio.sleep(1)

    async def _test_overlapping(self):
        """
        When client allows async, there's a chance we would send overlapping requests, with
          the second request starting before the first response is received. Make sure this
          works as expected.
        """

        reqRepPort = 7898

        serverState = Test_ZMQConnector.state()
        serverState.a = 1
        serverState.b = 'two'
        addrKey = 'testZMQConnector'
        server = ZMQConnectorServer(obj=serverState, reqRepPort=reqRepPort)
        await asyncio.sleep(1)
        logger.debug("Finished initializing server")
        client = ZMQConnectorClient(reqRepPort=reqRepPort, allowAsyncCalls=True)

        await asyncio.sleep(1)

        clientState = Test_ZMQConnector.state()
        clientState.c = 3

        futures: tp.List[asyncio.Future] = list()
        futures.append(client.callAsync_async('respondAfterDelay', delay=5., toReturn='call1'))
        futures.append(client.callAsync_async('respondAfterDelay', delay=2., toReturn='call2'))
        futures.append(client.get_async('a'))
        futures.append(client.get_async('b'))
        futures.append(client.set_async('c', clientState.c))

        logger.debug('Awaiting results from overlapping calls')
        (resp1, resp2, clientState.a, clientState.b, shouldBeNone) = await asyncio.gather(*futures)
        logger.debug('Results received')

        self.assertEqual(resp1, 'call1')
        self.assertEqual(resp2, 'call2')
        self.assertEqual(serverState.a, clientState.a)
        self.assertEqual(serverState.b, clientState.b)
        self.assertEqual(serverState.c, clientState.c)

        server.close()  # needed for successive tests
        del server
        await asyncio.sleep(1)

    def _test_large_sync(self, client, clientState):
        logger.debug('Client getting large state from server')
        clientState.c = client.get('c', raw=True)
        logger.debug('Done getting large state')
        client.set('d', clientState.d, raw=True)
        return clientState

    async def _test_large(self, doAsync):
        import os
        reqRepPort = 7899
        logger.debug('Initializing server state')
        serverState = Test_ZMQConnector.state()
        serverState.c = [os.urandom(2**27)]

        logger.debug('Initializing server connector')
        server = ZMQConnectorServer(obj=serverState, reqRepPort=reqRepPort)
        await asyncio.sleep(1)

        logger.debug('Initializing client connector')
        client = ZMQConnectorClient(reqRepPort=reqRepPort, allowAsyncCalls=doAsync)
        await asyncio.sleep(1)

        clientState = Test_ZMQConnector.state()
        clientState.d = [os.urandom(2 ** 27)]

        if doAsync:
            logger.debug('Client getting large state from server')
            clientState.c = await client.get_async('c', raw=True)
            logger.debug('Done getting large state')
            await client.set_async('d', clientState.d, raw=True)
        else:
            loop = asyncio.get_running_loop()
            clientState = await loop.run_in_executor(None, lambda: self._test_large_sync(client, clientState))

        self.assertEqual(serverState.c, clientState.c)

        self.assertEqual(serverState.d[0], clientState.d[0])

        server.close()
        client.close()

    def test_large_async(self):
        asyncio.get_event_loop().run_until_complete(self._test_large(doAsync=True))

    def test_large_sync(self):
        asyncio.get_event_loop().run_until_complete(self._test_large(doAsync=False))

    def test_basic_async(self):
        asyncio.get_event_loop().run_until_complete(self._test_basic(doAsync=True))

    def test_basic_sync(self):
        asyncio.get_event_loop().run_until_complete(self._test_basic(doAsync=False))

    def test_overlapping_async(self):
        asyncio.get_event_loop().run_until_complete(self._test_overlapping())






import asyncio
import attrs
import logging
import numpy as np
import pyigtl
import typing as tp
from typing import ClassVar

from NaviNIBS.Devices.ToolPositionsServer import ToolPositionsServer, TimestampedToolPosition


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


@attrs.define
class IGTLinkToolPositionsServer(ToolPositionsServer):
    """
    This acts as a central router for IGTLink position updates.

    It actually acts as a client connecting to a running Plus Server that is itself streaming tool positions.
    But this provides other clients a connection-agnostic async interface for receiving updates.
    """

    _type: ClassVar[str] = 'IGTLink'

    _igtlHostname: str = '127.0.0.1'
    _igtlPort: int = 18944
    _igtlClient: pyigtl.OpenIGTLinkClient = attrs.field(init=False)
    _igtlPollPeriod: float = 0.01

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        logger.info('Initializing IGTLink client')
        self._igtlClient = pyigtl.OpenIGTLinkClient(host=self._igtlHostname, port=self._igtlPort, start_now=False)

    async def run(self):
        logger.info('Starting IGTLink client')
        self._igtlClient.start()
        await asyncio.sleep(1.)  # give some time for client to connect
        logger.info('Starting IGTLink polling')
        while True:
            msgs = self._igtlClient.get_latest_messages()
            if len(msgs) == 0:
                logger.debug('No new IGTLink messages')
                await asyncio.sleep(self._igtlPollPeriod)
            else:
                logger.debug('{} new IGTLink messages'.format(len(msgs)))
                async with self._publishingLatestLock:  # delay publishing until after handling all these messages
                    for msg in msgs:
                        if msg.message_type == 'TRANSFORM':
                            transf = msg.matrix
                            if np.allclose(transf, np.eye(4)):
                                # plus sends identity when transforms are invalid
                                logger.debug('Transform for {} is invalid'.format(msg.device_name))
                                transf = None

                            position = TimestampedToolPosition(time=msg.timestamp,
                                                               transf=transf)
                            key = msg.device_name
                            if key.endswith('ToTracker'):
                                # strip 'ToTracker' suffix from device_name, assuming plus config is set to only send *ToTracker transforms
                                key = key[:-len('ToTracker')]

                            logger.debug(f'Transform for key {key}')
                            await self.recordNewPosition(key=key, position=position)
                        else:
                            logger.error('Unexpected message type: {}'.format(msg.message_type))
                            raise NotImplementedError()
                await asyncio.sleep(0)


if __name__ == '__main__':
    IGTLinkToolPositionsServer.createAndRun()


from __future__ import annotations

import attrs
import copy
from datetime import datetime
import nibabel as nib
import json
import logging
import numpy as np
import os
import pandas as pd
import pyvista as pv
import shutil
import tempfile
import typing as tp
from typing import ClassVar

from NaviNIBS.util.attrs import attrsAsDict
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.numpy import array_equalish, attrsWithNumpyAsDict, attrsWithNumpyFromDict, attrsOptionalNDArrayField

from NaviNIBS.Navigator.Model.GenericCollection import GenericCollection, GenericCollectionDictItem


logger = logging.getLogger(__name__)


FiducialCoord = np.ndarray
FiducialSet = tp.Dict[str, tp.Optional[FiducialCoord]]
Transform = np.ndarray
HeadPoint = np.ndarray


@attrs.define
class Fiducial(GenericCollectionDictItem[str]):
    _plannedCoord: tp.Optional[np.ndarray] = attrsOptionalNDArrayField()
    """
    Note that these (planned) coordinates are in MRI space.
    """
    _sampledCoords: tp.Optional[np.ndarray] = attrsOptionalNDArrayField()
    """
    Note that these (sampled) coordinates are in tracker space, not in MRI space. This is to not require modification when
    using other methods besides fiducials for defining tracker->MRI coordinate transform.
    
    Is of size Nx3, where N is the number of repeated samples of the same coordinate.
    """
    _sampledCoord: tp.Optional[np.ndarray] = attrsOptionalNDArrayField()
    """
    Is of size (3,). 
    
    Can be used without sampledCoords to specify single sampled coordinate. Or can be used with sampledCoords to override the default sampledCoords->sampledCoord aggregation (a simple mean) with some other result (e.g. a trimmed mean) while still preserving all original samples.
    
    Note that this is cleared whenever sampledCoords is set, to avoid persisting an out-of-date aggregated sample. 
    """

    _alignmentWeight: float = 1.
    """
    Can be used to specify a relative positive weight for how this fiducial will be incorporated
    into registration. 
    
    For example, if the NAS fiducial is given weight 100, and the LPA and RPA
    fiducials have default weights of 1, then alignment will always very closely match measured 
    NAS to planned NAS location, and allow larger deviations for LPA and RPA for the sake of 
    closely matching NAS.  
    """

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if self._sampledCoords is not None:
            if self._sampledCoords.ndim == 1:
                assert self._sampledCoords.shape[0] == 3
                # reshape from (3,) to (1, 3) to be consistent with size when there are multiple points
                self._sampledCoords = self._sampledCoords[np.newaxis, :]
            else:
                assert self._sampledCoords.ndim == 2
                assert self._sampledCoords.shape[1] == 3

        if self._sampledCoord is not None:
            assert self._sampledCoord.ndim == 1 and self._sampledCoord.shape[0] == 3

    @property
    def plannedCoord(self):
        return self._plannedCoord

    @plannedCoord.setter
    def plannedCoord(self, newCoord: tp.Optional[np.ndarray]):
        if array_equalish(newCoord, self._plannedCoord):
            return
        self.sigItemAboutToChange.emit(self._key, ['plannedCoord'])
        self._plannedCoord = newCoord
        self.sigItemChanged.emit(self._key, ['plannedCoord'])

    @property
    def sampledCoord(self):
        """
        Always returns single vector coordinate of size (3,), or None if the fiducial has not been sampled.
        If multiple repeats of the same fiducial are sampled (in sampledCoords), and no separate sampledCoord
        has been set, this returns an average of the samples.
        """
        if self._sampledCoord is not None:
            return self._sampledCoord
        elif self._sampledCoords is None:
            return None
        elif self._sampledCoords.shape[0] == 1:
            return self._sampledCoords.squeeze()
        else:
            return self._sampledCoords.mean(axis=0, keepdims=False)

    @sampledCoord.setter
    def sampledCoord(self, newCoord: tp.Optional[np.ndarray]):
        if array_equalish(newCoord, self._sampledCoord):
            return

        if newCoord is not None:
            assert newCoord.ndim == 1 and newCoord.shape[0] == 3
        changingAttrs = ['sampledCoord']
        if self._sampledCoords is None:
            changingAttrs.append('sampledCoords')
        self.sigItemAboutToChange.emit(self._key, changingAttrs)
        self._sampledCoord = newCoord
        self.sigItemChanged.emit(self._key, changingAttrs)

    @property
    def sampledCoords(self):
        if self._sampledCoords is not None:
            return self._sampledCoords
        elif self._sampledCoord is not None:
            return self._sampledCoord[np.newaxis, :]
        else:
            return None

    @sampledCoords.setter
    def sampledCoords(self, newCoords: tp.Optional[np.ndarray]):
        if array_equalish(newCoords, self._sampledCoords):
            if newCoords is None and self._sampledCoord is not None:
                self.sampledCoord = None
            return

        if newCoords is not None:
            assert newCoords.ndim == 2 and newCoords.shape[1] == 3
        changingAttrs = ['sampledCoords']
        if newCoords is not None or self._sampledCoord is not None:
            changingAttrs.append('sampledCoord')
        self.sigItemAboutToChange.emit(self._key, changingAttrs)
        self._sampledCoord = None  # reset whenever sampledCoords is set to avoid persisting an out-of-date aggregated sample
        self._sampledCoords = newCoords
        self.sigItemChanged.emit(self._key, changingAttrs)

    @property
    def alignmentWeight(self):
        return self._alignmentWeight

    @alignmentWeight.setter
    def alignmentWeight(self, newWeight: float):
        if newWeight == self._alignmentWeight:
            return
        self.sigItemAboutToChange.emit(self._key, ['alignmentWeight'])
        self._alignmentWeight = newWeight
        self.sigItemChanged.emit(self._key, ['alignmentWeight'])

    def asDict(self) -> dict[str, tp.Any]:
        npFields = ['plannedCoord', 'sampledCoord', 'sampledCoords']
        return attrsWithNumpyAsDict(self, npFields=npFields)

    @classmethod
    def fromDict(cls, d: dict[str, tp.Any]):
        npFields = ['plannedCoord', 'sampledCoord', 'sampledCoords']
        return attrsWithNumpyFromDict(cls, d, npFields=npFields)


@attrs.define
class Fiducials(GenericCollection[str, Fiducial]):
    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @property
    def plannedFiducials(self) -> dict[str, tp.Optional[np.ndarray]]:
        return {key: fid.plannedCoord for key, fid in self.items()}

    @property
    def sampledFiducials(self) -> dict[str, tp.Optional[np.ndarray]]:
        """
        Note that this returns a dict where each value is a `sampledCoord` (always None or shape (3,)),
        even if multiple samples exist for a given fiducial. For full access, use `allSampledFiducials`
        instead.
        """
        return {key: fid.sampledCoord for key, fid in self.items()}

    @property
    def allSampledFiducials(self) -> dict[str, tp.Optional[np.ndarray]]:
        return {key: fid.sampledCoords for key, fid in self.items()}

    @property
    def alignmentWeights(self) -> np.ndarray:
        return np.asarray([fid.alignmentWeight for fid in self.values()])

    @classmethod
    def fromList(cls, itemList: list[dict[str, tp.Any]]) -> Fiducials:
        items = {}
        for itemDict in itemList:
            items[itemDict['key']] = Fiducial.fromDict(itemDict)

        return cls(items=items)


@attrs.define
class HeadPoints:
    _headPoints: list[HeadPoint] = attrs.field(factory=list)
    """
    Note that these are coordinates are in tracker space, not in MRI space.
    """
    _alignmentWeights: np.ndarray | None = attrs.field(default=None)
    """
    Optional weights used for headpoint-based registration refinement, in format expected by 
    simpleicp's `rbp_observation_weights` argument (i.e. rot_x, rot_y, rot_z, t_x, t_y, t_z). 
    Can alternatively specify as a single scalar to apply the same weight to all terms, or as 
    two scalars to apply the first weight to all rotation term, the second weights to all
    translation terms.
    
    Note that values will be inverted (weightArg = 1 / weight) so that higher input values 
    correspond to greater weighting of head points, rather than greater weighting of fiducials.
    
    Note that these weights are different in definition and usage than `Fiducial.alignmentWeight`.
    """

    sigHeadpointsAboutToChange: Signal = attrs.field(init=False, factory=lambda: Signal((list[int],)))
    """
    This signal includes list of indices of headpoints about to change.
    """

    sigHeadpointsChanged: Signal = attrs.field(init=False, factory=lambda: Signal((list[int],)))
    """"
    This signal includes list of indices of headpoints that changed.
    """

    sigAttribsAboutToChange: Signal = attrs.field(init=False, factory=lambda: Signal((dict[str],)))
    """
    This signal includes keys and incoming values of attributes (besides main headPoints list) about to change.
    """

    sigAttribsChanged: Signal = attrs.field(init=False, factory=lambda: Signal((list[str],)))
    """
    This signal includes keys of attributes (besides main headPoints list) that changed.
    """

    def __attrs_post_init__(self):
        pass

    @property
    def alignmentWeights(self):
        return self._alignmentWeights

    @alignmentWeights.setter
    def alignmentWeights(self, newWeights: tp.Optional[np.ndarray]):
        if array_equalish(newWeights, self._alignmentWeights):
            return
        if newWeights is not None:
            assert isinstance(newWeights, np.ndarray)
            assert len(newWeights) in (1, 2, 6)
        self.sigAttribsAboutToChange.emit(dict(alignmentWeights=newWeights))
        self._alignmentWeights = newWeights
        self.sigAttribsChanged.emit(['alignmentWeights'])

    def append(self, point: HeadPoint):
        index = len(self)
        self.sigHeadpointsAboutToChange.emit([index])
        self._headPoints.append(point)
        self.sigHeadpointsChanged.emit([index])

    def extend(self, points: np.ndarray):
        assert points.shape[1] == 3
        indices = len(self) + np.arange(points.shape[0])
        self.sigHeadpointsAboutToChange.emit(indices)
        self._headPoints.extend(points)
        self.sigHeadpointsChanged.emit(indices)

    def remove(self, indexOrIndices: int | list[int]):
        if isinstance(indexOrIndices, int):
            indexOrIndices = [indexOrIndices]

        indexOrIndices = sorted(indexOrIndices)

        changedIndices = list(range(indexOrIndices[0], len(self)));
        self.sigHeadpointsAboutToChange.emit(changedIndices)
        for index in reversed(indexOrIndices):
            del self._headPoints[index]
        self.sigHeadpointsChanged.emit(changedIndices)

    def replace(self, points: np.ndarray):
        """
        Clear all previous points, set to new points.
        """
        indices = list(range(0, max(len(points), len(self))))
        self.sigHeadpointsAboutToChange.emit(indices)
        self._headPoints.clear()
        with self.sigHeadpointsAboutToChange.blocked(), self.sigHeadpointsChanged.blocked():
            self.extend(points)
        self.sigHeadpointsChanged.emit(indices)

    def clear(self):
        indices = list(range(0, len(self)))
        self.sigHeadpointsAboutToChange.emit(indices)
        self._headPoints.clear()
        self.sigHeadpointsChanged.emit(indices)

    def __len__(self):
        return len(self._headPoints)

    def __iter__(self):
        return iter(self._headPoints)

    def __getitem__(self, index):
        return self._headPoints[index]

    def asNDArray(self):
        return self._headPoints

    def asList(self):
        return [headPt.tolist() for headPt in self._headPoints]

    @classmethod
    def fromList(cls, l: list[tuple[float, float, float]]):
        return cls(headPoints=[np.asarray(headPt) for headPt in l])


@attrs.define()
class SubjectRegistration:

    _fiducials: Fiducials = attrs.field(factory=Fiducials)

    _sampledHeadPoints: HeadPoints = attrs.field(factory=HeadPoints)  # in head tracker space
    _trackerToMRITransf: tp.Optional[Transform] = None

    _fiducialsHistory: tp.Dict[str, Fiducials] = attrs.field(factory=dict)
    _trackerToMRITransfHistory: tp.Dict[str, tp.Optional[Transform]] = attrs.field(factory=dict)

    sigTrackerToMRITransfAboutToChange: Signal = attrs.field(init=False, factory=Signal)
    sigTrackerToMRITransfChanged: Signal = attrs.field(init=False, factory=Signal)

    def __attrs_post_init__(self):

        # make sure histories are up to date with current values
        if len(self._fiducials) > 0 and (len(self._fiducialsHistory) == 0 or
                                         not self.fiducialsEqual(list(self._fiducialsHistory.values())[-1], self._fiducials)):
            self._saveFiducialsToHistory()

        self._fiducials.sigItemsChanged.connect(self._onFiducialsChanged)

        if self._trackerToMRITransf is not None:
            if len(self._trackerToMRITransfHistory) == 0 or not array_equalish(list(self._trackerToMRITransfHistory.values())[-1], self._trackerToMRITransf):
                self._trackerToMRITransfHistory[self._getTimestampStr()] = self._trackerToMRITransf.copy()

    def _saveFiducialsToHistory(self):
        # TODO: make this more efficient
        self._fiducialsHistory[self._getTimestampStr()] = Fiducials.fromList(copy.deepcopy(self._fiducials.asList()))

    def _onFiducialsChanged(self, keys, attribs: tp.Optional[list[str]] = None):
        self._saveFiducialsToHistory()

    @property
    def fiducials(self):
        return self._fiducials

    @property
    def fiducialsHistory(self):
        """
        Result should not be modified
        """
        return self._fiducialsHistory

    @property
    def hasMinimumPlannedFiducials(self) -> bool:
        numFiducialsSet = 0
        for fid in self._fiducials.values():
            if fid.plannedCoord is not None:
                numFiducialsSet += 1
        return numFiducialsSet >= 3

    @property
    def hasMinimumSampledFiducials(self) -> bool:
        numFiducialsSet = 0
        for fid in self._fiducials.values():
            if fid.sampledCoord is not None and fid.plannedCoord is not None:
                numFiducialsSet += 1
        return numFiducialsSet >= 3

    @property
    def sampledHeadPoints(self) -> HeadPoints:
        return self._sampledHeadPoints
    
    @property
    def trackerToMRITransf(self):
        return self._trackerToMRITransf

    @trackerToMRITransf.setter
    def trackerToMRITransf(self, newTransf: tp.Optional[Transform]):
        if array_equalish(self._trackerToMRITransf, newTransf):
            logger.debug('No change in trackerToMRITransf, returning')
            return

        # TODO: do validation of newTransf

        logger.info('Set trackerToMRITransf to {}'.format(newTransf))

        self.sigTrackerToMRITransfAboutToChange.emit()

        self._trackerToMRITransf = newTransf

        self._trackerToMRITransfHistory[self._getTimestampStr()] = None if newTransf is None else newTransf.copy()

        self.sigTrackerToMRITransfChanged.emit()

    @property
    def trackerToMRITransfHistory(self):
        """
        Result should not be modified
        """
        return self._trackerToMRITransfHistory

    @property
    def isRegistered(self):
        return self.trackerToMRITransf is not None

    @property
    def approxHeadCenter(self) -> tp.Optional[np.ndarray]:
        lpa = self.fiducials.get('LPA', None)
        rpa = self.fiducials.get('RPA', None)
        if lpa is not None and rpa is not None and lpa.plannedCoord is not None and rpa.plannedCoord is not None:
            center = (lpa + rpa) / 2
        else:
            logger.warning('Insufficient information for determining approximate header center')
            # TODO: implement more general method of estimating center, e.g. more variety of LPA/RPA naming, name-agnostic averaging, etc.
            center = None
        return center

    def asDict(self) -> tp.Dict[str, tp.Any]:
        d = dict()

        d['fiducials'] = self._fiducials.asList()

        d['fiducialsHistory'] = [dict(time=key, fiducials=val.asList()) for key, val in self._fiducialsHistory.items()]

        d['sampledHeadPoints'] = self._sampledHeadPoints.asList()

        for key in ('fiducials', 'fiducialsHistory', 'sampledHeadPoints'):
            if len(d[key]) == 0:
                del d[key]  # don't include in output if it's empty anyways

        if self._sampledHeadPoints.alignmentWeights is not None:
            d['headPointAlignmentWeights'] = self._sampledHeadPoints.alignmentWeights.tolist()

        if self._trackerToMRITransf is not None:
            d['trackerToMRITransf'] = self._trackerToMRITransf.tolist()
        if len(self._trackerToMRITransfHistory) > 0:
            d['trackerToMRITransfHistory'] = [dict(time=key, trackerToMRITransf=val.tolist() if val is not None else None) for key, val in self._trackerToMRITransfHistory.items()]

        return d

    @classmethod
    def fromDict(cls, d: tp.Dict[str, tp.Any]) -> SubjectRegistration:
        # TODO: validate against schema

        # note: input dict is modified for arg conversion below

        # convert types as needed (e.g. if coming from deserialized json)

        def convertTransf(transf: tp.List[tp.List[float, float, float]]) -> np.ndarray:
            return np.asarray(transf)

        def convertHistoryListToDict(historyList: tp.List[tp.Dict[str, tp.Any]], field: str, entryConverter: tp.Callable) -> tp.Dict[str, tp.Any]:
            historyDict = {}
            for entry in historyList:
                timeStr = entry['time']
                historyDict[timeStr] = entryConverter(entry[field])
            return historyDict

        d['fiducials'] = Fiducials.fromList(d['fiducials'])

        if 'fiducialsHistory' in d:
            d['fiducialsHistory'] = {entry['time']: Fiducials.fromList(entry['fiducials'])
                                     for entry in d['fiducialsHistory']}

        if 'sampledHeadPoints' in d:
            d['sampledHeadPoints'] = HeadPoints.fromList(d['sampledHeadPoints'])

        if 'headPointAlignmentWeights' in d:
            d['sampledHeadPoints'].alignmentWeights = np.asarray(d.pop('headPointAlignmentWeights'))

        if 'trackerToMRITransf' in d:
            d['trackerToMRITransf'] = convertTransf(d['trackerToMRITransf'])

        if 'trackerToMRITransfHistory' in d:
            d['trackerToMRITransfHistory'] = convertHistoryListToDict(d['trackerToMRITransfHistory'],
                                                                      field='trackerToMRITransf',
                                                                      entryConverter=convertTransf)

        return cls(**d)

    @staticmethod
    def _getTimestampStr():
        return datetime.today().strftime('%y%m%d%H%M%S.%f')

    @staticmethod
    def getDatetimeFromTimestamp(ts: str) -> datetime:
        return datetime.strptime(ts, '%y%m%d%H%M%S.%f')

    @staticmethod
    def fiducialsEqual(fidsA: Fiducials, fidsB: Fiducials) -> bool:
        return json.dumps(fidsA.asList()) == json.dumps(fidsB.asList())



from __future__ import annotations

import asyncio
import attrs
import logging
import numpy as np

import pandas as pd
import qtawesome as qta
from qtpy import QtWidgets, QtGui, QtCore
import typing as tp

from .NavigationView import NavigationView, TargetingCrosshairsView, SinglePlotterNavigationView
from NaviNIBS.Navigator.TargetingCoordinator import TargetingCoordinator
from NaviNIBS.Navigator.GUI.Widgets.CollectionTableWidget import SamplesTableWidget
from NaviNIBS.Navigator.GUI.Widgets.CollectionTableWidget import TargetsTableWidget
from NaviNIBS.Navigator.GUI.Widgets.TrackingStatusWidget import TrackingStatusWidget
from NaviNIBS.Navigator.GUI.ViewPanels.MainViewPanelWithDockWidgets import MainViewPanelWithDockWidgets
from NaviNIBS.Navigator.Model.Session import Session
from NaviNIBS.Navigator.Model.Tools import CalibrationPlate, Pointer
from NaviNIBS.Navigator.Model.Triggering import TriggerReceiver, TriggerEvent
from NaviNIBS.Navigator.Model.Samples import Sample
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.CoilOrientations import PoseMetricCalculator
from NaviNIBS.util.GUI.Dock import Dock
from NaviNIBS.util.GUI.Icons import getIcon
from NaviNIBS.util.GUI.QScrollContainer import QScrollContainer
from NaviNIBS.util.GUI.QFlowLayout import QFlowLayout
from NaviNIBS.util.pyvista import DefaultBackgroundPlotter, RemotePlotterProxy

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

Transform = np.ndarray


@attrs.define
class _PoseMetricGroup:
    """
    Grouped information for displaying pose metrics in a widget.

    Set up to allow initial specification of title, container, and calculator key str, then finish initialization when calculator is set later.

    """
    _title: str
    _container: QtWidgets.QWidget
    _scroll: QScrollContainer = attrs.field(init=False)
    _calculatorKey: str
    _calculator: tp.Optional[PoseMetricCalculator] = None
    _indicators: dict[str, QtWidgets.QLabel] = attrs.field(init=False, factory=dict)
    _hasInitialized: bool = attrs.field(init=False, default=False)
    _needsUpdateEvent: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)
    _updateRateLimit: float = 2  # in Hz

    def __attrs_post_init__(self):
        asyncio.create_task(asyncTryAndLogExceptionOnError(self._loop_keepUpdated))

    @property
    def calculatorKey(self):
        return self._calculatorKey

    @property
    def calculator(self):
        return self._calculator

    @calculator.setter
    def calculator(self, calculator: PoseMetricCalculator):
        if calculator is self._calculator:
            return

        assert isinstance(calculator, PoseMetricCalculator)

        self._calculator = calculator

        poseMetrics = [poseMetric for poseMetric in self._calculator.supportedMetrics if poseMetric.doShowByDefault]

        if not self._hasInitialized:
            innerLayout = QtWidgets.QFormLayout()
            innerLayout.setContentsMargins(0, 0, 0, 0)
            self._scroll = QScrollContainer(innerContainerLayout=innerLayout)
            self._container.setLayout(QtWidgets.QVBoxLayout())
            self._container.layout().setContentsMargins(0, 0, 0, 0)
            self._container.layout().addWidget(self._scroll.scrollArea)
            self._container.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)

            for poseMetric in poseMetrics:
                wdgt = QtWidgets.QLabel(f"NaN{poseMetric.units}")
                self._scroll.innerContainerLayout.addRow(poseMetric.label, wdgt)
                self._indicators[poseMetric.label] = wdgt

            self._hasInitialized = True
        else:
            assert len(poseMetrics) == self._container.layout().rowCount()
            assert len(poseMetrics) == len(self._indicators)
            assert all(poseMetric.label in self._indicators for poseMetric in poseMetrics)

        self._calculator.sigCacheReset.connect(self._updateSoon)

        self._update()

    def _updateSoon(self):
        self._needsUpdateEvent.set()

    async def _loop_keepUpdated(self):
        while True:
            await self._needsUpdateEvent.wait()
            self._update()
            await asyncio.sleep(1/self._updateRateLimit)

    def _update(self):
        logger.debug(f'Updating pose metric indicators for {self._title}')
        self._needsUpdateEvent.clear()
        poseMetrics = [poseMetric for poseMetric in self._calculator.supportedMetrics if poseMetric.doShowByDefault]
        for poseMetric in poseMetrics:
            wdgt = self._indicators[poseMetric.label]
            val = poseMetric.getter()
            wdgt.setText(f"{val:.1f}{poseMetric.units}")
        logger.debug(f'Done updating pose metric indicators for {self._title}')


@attrs.define
class BackgroundSamplePoseMetadataSetter:
    _session: Session | None = attrs.field(init=False, default=None, repr=False)
    _pendingSampleKeys: list[str] = attrs.field(init=False, factory=list)
    _needsUpdateEvent: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)
    _calculator: PoseMetricCalculator | None = attrs.field(init=False, default=None)

    def __attrs_post_init__(self):
        asyncio.create_task(asyncTryAndLogExceptionOnError(self._loop_keepUpdated))

    async def _loop_keepUpdated(self):
        while True:
            await self._needsUpdateEvent.wait()
            self._needsUpdateEvent.clear()
            if self._session is None:
                continue

            if len(self._pendingSampleKeys) == 0:
                continue

            sampleKey = self._pendingSampleKeys.pop(0)

            if len(self._pendingSampleKeys) > 0:
                self._needsUpdateEvent.set()

            if sampleKey not in self.session.samples:
                # sample was presumably deleted while pending
                continue

            sample = self.session.samples[sampleKey]

            if self._calculator is None:
                self._calculator = PoseMetricCalculator(
                    session=self.session,
                    sample=sample
                )
            else:
                self._calculator.sample = sample

            with sample.changingMetadata():
                for metric in self._calculator.supportedMetrics:
                    if not metric.doShowByDefault:
                        continue
                    sample.metadata[metric.label] = self._calculator.getValueForMetric(metric.label)

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, newSes: Session | None):
        if self._session is newSes:
            return
        if self._session is not None:
            raise NotImplementedError  # TODO: disconnect from signals, update self._calculator.session, etc.
        else:
            assert self._calculator is None  # will be given session ref when instantiated
        self._session = newSes
        self._session.samples.sigItemKeyChanged.connect(self._onSampleKeyChanged)
        self._needsUpdateEvent.set()

    def queueSamples(self, sampleKeys: list[str]):
        self._pendingSampleKeys.extend([key for key in sampleKeys if key not in self._pendingSampleKeys])
        self._needsUpdateEvent.set()

    def _onSampleKeyChanged(self, oldKey: str, newKey: str):
        if oldKey in self._pendingSampleKeys:
            if newKey not in self._pendingSampleKeys:
                self._pendingSampleKeys.append(newKey)
                self._needsUpdateEvent.set()


@attrs.define
class NavigatePanel(MainViewPanelWithDockWidgets):
    _key: str = 'Navigate'
    _icon: QtGui.QIcon = attrs.field(init=False, factory=lambda: getIcon('mdi6.head-flash'))
    _trackingStatusWdgt: TrackingStatusWidget = attrs.field(init=False, repr=False)
    _targetsTableWdgt: TargetsTableWidget = attrs.field(init=False, repr=False)
    _poseMetricGroups: list[_PoseMetricGroup] = attrs.field(init=False, factory=list, repr=False)
    _samplesTableWdgt: SamplesTableWidget = attrs.field(init=False, repr=False)
    _sampleBtn: QtWidgets.QPushButton = attrs.field(init=False, repr=False)
    _hideAllSamplesBtn: QtWidgets.QPushButton = attrs.field(init=False, repr=False)
    _sampleToTargetBtn: QtWidgets.QPushButton = attrs.field(init=False, repr=False)
    _views: dict[str, NavigationView] = attrs.field(init=False, factory=dict, repr=False)

    _coordinator: TargetingCoordinator = attrs.field(init=False, repr=False)
    _triggerReceiver: TriggerReceiver | None = attrs.field(init=False, default=None, repr=False)
    _backgroundSamplePoseMetadataSetter: BackgroundSamplePoseMetadataSetter = attrs.field(
        init=False,
        factory=BackgroundSamplePoseMetadataSetter,
        repr=False
    )

    _hasInitializedNavigationViews: bool = attrs.field(init=False, default=False)

    finishedAsyncInit: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def canBeEnabled(self) -> tuple[bool, str | None]:
        if self.session is None:
            return False, 'No session set'
        if not self.session.MRI.isSet:
            return False, 'No MRI set'
        if not self.session.headModel.isSet:
            return False, 'No head model set'
        if self.session.tools.subjectTracker is None:
            return False, 'No active subject tracker configured'
        if not self.session.subjectRegistration.isRegistered:
            return False, 'Subject not registered'
        return True, None

    def _finishInitialization(self):
        super()._finishInitialization()

        dock, container = self._createDockWidget(
            title='Tools tracking status',
        )
        self._trackingStatusWdgt = TrackingStatusWidget(wdgt=container,
                                                        hideToolTypes=[CalibrationPlate, Pointer])
        self._wdgt.addDock(dock, position='left')
        dock.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)

        self._targetsTableWdgt = TargetsTableWidget()
        self._targetsTableWdgt.sigCurrentItemChanged.connect(self._onCurrentTargetChanged)
        self._targetsTableWdgt.wdgt.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        container.layout().addWidget(self._targetsTableWdgt.wdgt)

        dock, container = self._createDockWidget(
            title='Targets',
            widget=self._targetsTableWdgt.wdgt
        )
        dock.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)
        self._wdgt.addDock(dock, position='bottom')

        poseGroupDicts = [
            dict(title='Current pose metrics', calculatorKey='currentPoseMetrics'),
            dict(title='Selected sample metrics', calculatorKey='currentSamplePoseMetrics')
        ]

        prevPoseDock = None
        for poseGroupDict in poseGroupDicts:
            dock, container = self._createDockWidget(
                title=poseGroupDict['title'],
            )
            dock.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)

            poseGroup = _PoseMetricGroup(
                title=poseGroupDict['title'],
                container=container,
                calculatorKey=poseGroupDict['calculatorKey']
            )

            if prevPoseDock is None:
                self._wdgt.addDock(dock, position='bottom')
            else:
                self._wdgt.addDock(dock, position='below', relativeTo=prevPoseDock)
            prevPoseDock = dock

            self._poseMetricGroups.append(poseGroup)

        dock, container = self._createDockWidget(
            title='Samples',
            layout=QtWidgets.QVBoxLayout())
        dock.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        self._wdgt.addDock(dock, position='bottom')

        btnContainer = QtWidgets.QWidget()
        btnContainerLayout = QFlowLayout()
        btnContainerLayout.setContentsMargins(0, 0, 0, 0)
        btnContainer.setLayout(btnContainerLayout)
        container.layout().addWidget(btnContainer)

        btn = QtWidgets.QPushButton('Add a sample')
        btn.clicked.connect(self._onSampleBtnClicked)
        btnContainerLayout.addWidget(btn)
        self._sampleBtn = btn
        # TODO: change color to warning indicator when coil or tracker are not visible

        btn = QtWidgets.QPushButton('Sample ⇒ Target')
        btn.clicked.connect(self._onSampleToTargetBtnClicked)
        btnContainerLayout.addWidget(btn)
        btn.setEnabled(False)
        self._sampleToTargetBtn = btn
        # TODO: only enable when one or more samples are selected

        btn = QtWidgets.QPushButton('Hide all samples')
        btn.clicked.connect(self._onHideAllSamplesBtnClicked)
        btnContainerLayout.addWidget(btn)
        self._hideAllSamplesBtn = btn

        # TODO: add a 'Create target from pose' button (but clearly separate, maybe in different panel, from 'Create target from sample' button)

        self._samplesTableWdgt = SamplesTableWidget()
        self._samplesTableWdgt.sigCurrentItemChanged.connect(self._onCurrentSampleChanged)
        self._samplesTableWdgt.sigSelectionChanged.connect(self._onSelectedSamplesChanged)
        self._samplesTableWdgt.wdgt.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        container.layout().addWidget(self._samplesTableWdgt.wdgt)

        self._triggerReceiver = TriggerReceiver(
            key=self._key,
            #minTimeBetweenEvents=1.1,  # TODO: make this GUI-configurable and config-file-configurable
        )
        self._triggerReceiver.sigTriggered.connect(self._onReceivedTrigger)

        if self.session is not None:
            self._onPanelInitializedAndSessionSet()

        asyncio.create_task(asyncTryAndLogExceptionOnError(self._finishInitialization_async))

    async def _finishInitialization_async(self):
        for view in self._views.values():
            if isinstance(view, SinglePlotterNavigationView):
                await view.finishedAsyncInit.wait()

        self.finishedAsyncInit.set()

    def _onSessionSet(self):
        super()._onSessionSet()

        if self._hasInitialized:
            self._onPanelInitializedAndSessionSet()

    def _onPanelInitializedAndSessionSet(self):
        logger.debug(f'{self.__class__.__name__} onPanelInitializedAndSessionSet')
        self._trackingStatusWdgt.session = self.session
        self._coordinator = TargetingCoordinator.getSingleton(session=self.session)
        self._coordinator.sigCurrentTargetChanged.connect(lambda: self._onCurrentTargetChanged(self._coordinator.currentTargetKey))
        self._coordinator.sigCurrentSampleChanged.connect(lambda: self._onCurrentSampleChanged(self._coordinator.currentSampleKey))
        self._targetsTableWdgt.session = self.session
        self._samplesTableWdgt.session = self.session
        self._backgroundSamplePoseMetadataSetter.session = self.session

        self.session.samples.sigItemsChanged.connect(self._onSamplesChanged)

        # now that coordinator is available, finish initializing pose metric groups
        for poseGroup in self._poseMetricGroups:
            poseGroup.calculator = getattr(self._coordinator, poseGroup.calculatorKey)

        if not self._hasInitializedNavigationViews:
            self._initializeDefaultViews()  # TODO: only do this if not restoring from previously saved config

        # register that we want to receive sample triggers when visible
        self.session.triggerSources.triggerRouter.registerReceiver(self._triggerReceiver)
        if self._isShown:
            self.session.triggerSources.triggerRouter.subscribeToTrigger(
                receiver=self._triggerReceiver, triggerKey='sample', exclusive=True)
            self.session.triggerSources.triggerRouter.subscribeToTrigger(
                receiver=self._triggerReceiver, triggerKey='pulse', exclusive=True)

        self.restoreLayoutIfAvailable()

    def _onCurrentTargetChanged(self, newTargetKey: str | None):
        """
        Called when targetTreeWdgt selection or coordinator currentTarget changes, NOT when attributes of currently selected target change
        """
        if self._hasInitialized:
            logger.debug(f'Current target key changed to {newTargetKey}')
            self._coordinator.currentTargetKey = newTargetKey
            self._targetsTableWdgt.currentCollectionItemKey = newTargetKey

    def _onPanelShown(self):
        super()._onPanelShown()
        if self.session is not None and self._triggerReceiver is not None:
            self.session.triggerSources.triggerRouter.subscribeToTrigger(receiver=self._triggerReceiver, triggerKey='sample', exclusive=True)

        for view in self._views.values():
            if isinstance(view, SinglePlotterNavigationView):
                if not isinstance(view.plotter, RemotePlotterProxy) or view.plotter.isReadyEvent.is_set():
                    view.plotter.resumeRendering()

    def _onPanelHidden(self):
        super()._onPanelHidden()
        if self._hasInitialized and self.session is not None:
            self.session.triggerSources.triggerRouter.unsubscribeFromTrigger(receiver=self._triggerReceiver, triggerKey='sample', exclusive=True)

            for view in self._views.values():
                if isinstance(view, SinglePlotterNavigationView):
                    if not isinstance(view.plotter, RemotePlotterProxy) or view.plotter.isReadyEvent.is_set():
                        with view.plotter.allowNonblockingCalls():
                            view.plotter.pauseRendering()

    def _onSamplesChanged(self, changesKeys: list[str], changedAttribs: tp.Optional[list[str]] = None):
        if changedAttribs is None or 'isVisible' in changedAttribs:
            self._onSelectedSamplesChanged(self._samplesTableWdgt.selectedCollectionItemKeys)

    def _onCurrentSampleChanged(self, newSampleKey: str):
        """
        Called when sampleTreeWdgt selection or coordinator currentSample changes, NOT when attributes of currently selected sample change
        """
        if self._hasInitialized:
            self._coordinator.currentSampleKey = newSampleKey

    def _onSelectedSamplesChanged(self, selectedKeys: list[str]):
        if self._hasInitialized:
            numSelectedHidden = sum(not self.session.samples[key].isVisible for key in selectedKeys)

            numSelectedVisible = len(selectedKeys) - numSelectedHidden

            self._sampleToTargetBtn.setEnabled(len(selectedKeys) > 0)

    def _onSampleBtnClicked(self, _):
        self._recordSample(timestamp=pd.Timestamp.now())

    def _onReceivedTrigger(self, triggerEvt: TriggerEvent):
        self._recordSample(timestamp=triggerEvt.time,
                           metadata=triggerEvt.metadata)

    def _recordSample(self, timestamp: tp.Optional[pd.Timestamp], metadata: tp.Optional[dict[str, tp.Any]] = None):
        sampleKey = self.session.samples.getUniqueSampleKey(timestamp=timestamp)
        coilToMRITransf = self._coordinator.currentCoilToMRITransform  # may be None if missing a tracker, etc.

        if abs(timestamp - pd.Timestamp.now()).total_seconds() > 10:
            # We are getting "old" triggers or lagging for other reasons. Mark orientation as invalid
            logger.warning('Requested sample time is far from current time. Unable to get up-to-date orientation information')
            coilToMRITransf = None

        sample = Sample(
            key=sampleKey,
            timestamp=timestamp,
            coilToMRITransf=coilToMRITransf,
            targetKey=self._coordinator.currentTargetKey,
            coilKey=self._coordinator.activeCoilKey,
            metadata=metadata if metadata is not None else {}
        )

        logger.info(f'Manually recording a sample: {sample}')

        self.session.samples.addItem(sample)

        self._backgroundSamplePoseMetadataSetter.queueSamples([sample.key])

        logger.debug(f'Manually recorded a sample: {sample}')

    def _onSampleToTargetBtnClicked(self, _):
        newTarget = self._coordinator.createTargetFromCurrentSample(doAddToSession=True)
        self._targetsTableWdgt.currentCollectionItemKey = newTarget.key

    def _onShowSampleBtnClicked(self, _):
        selKeys = self._samplesTableWdgt.selectedCollectionItemKeys
        self.session.samples.setAttribForItems(selKeys, dict(isVisible=[True for key in selKeys]))

    def _onHideAllSamplesBtnClicked(self, _):
        self.session.samples.setWhichSamplesVisible([])

    def _onHideSampleBtnClicked(self, _):
        selKeys = self._samplesTableWdgt.selectedCollectionItemKeys
        self.session.samples.setAttribForItems(selKeys, dict(isVisible=[False for key in selKeys]))

    def _initializeDefaultViews(self):
        if self._hasInitializedNavigationViews:
            assert len(self._views) > 0
            raise NotImplementedError()  # TODO: clear any previous views from dock and self._views

        self.addView(key='Crosshairs', View='TargetingCrosshairs')
        self.addView(key='Crosshairs-X', View='TargetingCrosshairs-X')
        self.addView(key='Crosshairs-Y', View='TargetingCrosshairs-Y')

        # TODO: set up other default views

        self._hasInitializedNavigationViews = True

    def addView(self, key: str, View: str | tp.Callable[[...], NavigationView],
                position: str | None = None,
                positionRelativeTo: Dock | None = None,
                **kwargs):

        # TODO: maybe add optional code here to generate unique key using input as a base key
        assert key not in self._views

        if isinstance(View, str):
            match View:
                case 'TargetingCrosshairs':
                    View = TargetingCrosshairsView
                    kwargs.setdefault('title', 'Crosshairs (Z)')
                    kwargs.setdefault('doShowHandleAngleError', True)
                    kwargs.setdefault('cameraDist', 100)
                case 'TargetingCrosshairs-X':
                    View = TargetingCrosshairsView
                    kwargs.setdefault('title', 'Crosshairs (X)')
                    kwargs.setdefault('doParallelProjection', True)
                    kwargs.setdefault('alignCameraTo', 'coil+X')
                    kwargs.setdefault('alignCameraOffset', (0, 0, -20))
                    kwargs.setdefault('cameraDist', 75)
                    kwargs.setdefault('doShowSkinSurf', True)
                    kwargs.setdefault('doShowTargetTangentialAngleError', True)
                    kwargs.setdefault('doShowScalpTangentialAngleError', True)
                case 'TargetingCrosshairs-Y':
                    kwargs.setdefault('title', 'Crosshairs (Y)')
                    View = TargetingCrosshairsView
                    kwargs.setdefault('doParallelProjection', True)
                    kwargs.setdefault('alignCameraTo', 'coil-Y')
                    kwargs.setdefault('alignCameraOffset', (0, 0, -20))
                    kwargs.setdefault('cameraDist', 75)
                    kwargs.setdefault('doShowSkinSurf', True)
                    kwargs.setdefault('doShowTargetTangentialAngleError', True)
                    kwargs.setdefault('doShowScalpTangentialAngleError', True)

                case _:
                    raise NotImplementedError('Unexpected viewType: {}'.format(View))

        view: NavigationView = View(key=key, dockKeyPrefix=self._key, coordinator=self._coordinator, **kwargs)
        view.wdgt.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)

        assert view.key not in self._views
        self._views[view.key] = view

        if key == 'Crosshairs-Y':
            # TODO: use position and positionRelativeTo if either/both specified
            self._wdgt.addDock(view.dock,
                               position='right',
                               relativeTo=self._views['Crosshairs-X'].dock)
        elif key == 'Crosshairs-X':
            # TODO: use position and positionRelativeTo if either/both specified
            self._wdgt.addDock(view.dock,
                               position='bottom',
                               relativeTo=self._views['Crosshairs'].dock)
        else:
            if position is None:
                position = 'right'
            self._wdgt.addDock(view.dock,
                               position=position,
                               relativeTo=positionRelativeTo)



import typing as tp
import sys
import traceback
import logging


def exceptionToStr(e: Exception) -> str:

    # from https://stackoverflow.com/a/49613561
    ex_type, ex_value, ex_traceback = sys.exc_info()

    # Extract unformatter stack traces as tuples
    trace_back = traceback.extract_tb(ex_traceback)

    # Format stacktrace
    stack_trace = ''

    eStr = ''

    for trace in trace_back:
        stack_trace += "File : %s , Line : %d, Func.Name : %s, Message : %s\n" % (trace[0], trace[1], trace[2], trace[3])

    if ex_type is not None:
        eStr += "Exception type : %s\n" % ex_type.__name__
    eStr += "Exception message : %s\n" % ex_value
    eStr += "Stack trace : %s\n" % stack_trace

    return eStr


def makeStrUnique(baseStr: str, existingStrs: tp.Iterable[str], delimiter: str = '_') -> str:
    count = 1
    uniqueStr = baseStr

    if delimiter in baseStr:
        try:
            prevNum = int(baseStr[baseStr.rindex(delimiter)+1:])
        except ValueError as e:
            pass
        else:
            baseStr = baseStr[:baseStr.rindex(delimiter)]
            count = prevNum

    while uniqueStr in existingStrs:
        count += 1
        uniqueStr = '{}{}{}'.format(baseStr, delimiter, count)

    return uniqueStr


class classproperty:
    """
    Adapted from https://stackoverflow.com/a/13624858
    """
    def __init__(self, fget):
        self.fget = fget

    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)



from __future__ import annotations

import collections

import attrs
from abc import ABC
from collections.abc import Sequence, Mapping, Iterable
import logging
import typing as tp

from NaviNIBS.util.attrs import attrsAsDict
from NaviNIBS.util.numpy import array_equalish
from NaviNIBS.util.Signaler import Signal

logger = logging.getLogger(__name__)


K = tp.TypeVar('K', int, str)  # collection item key type


@attrs.define(slots=False)
class GenericCollectionDictItem(ABC, tp.Generic[K]):

    _key: K

    sigKeyAboutToChange: Signal = attrs.field(init=False, repr=False, eq=False,
                                              factory=lambda: Signal((K, K)))  # includes old key, new key
    sigKeyChanged: Signal = attrs.field(init=False, repr=False, eq=False,
                                        factory=lambda: Signal((K, K)))  # includes old key, new key

    sigItemAboutToChange: Signal = attrs.field(init=False, repr=False, eq=False,
                                               factory=lambda: Signal((K, tp.Optional[tp.List[str]])))
    """
    This signal includes the key of the item, and optionally a list of keys of attributes about to change;
    if second arg is None, all attributes should be assumed to be about to change.

    Not emitted when key changed (use sigKeyAboutToChange instead!)
    """
    sigItemChanged: Signal = attrs.field(init=False, repr=False, eq=False,
                                         factory=lambda: Signal((K, tp.Optional[list[str]])))
    """
    This signal includes the key of the item, and optionally a list of keys of changed attributes;  
    if second arg is None, all attributes should be assumed to have changed.

    Not emitted when key changed (use sigKeyChanged instead!)
    """

    def __attrs_post_init__(self):
        pass

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, newKey: K):
        if self._key == newKey:
            return
        prevKey = self._key
        self.sigKeyAboutToChange.emit(prevKey, newKey)
        self._key = newKey
        self.sigKeyChanged.emit(prevKey, newKey)

    def asDict(self) -> dict[str, tp.Any]:
        return attrsAsDict(self)

    @classmethod
    def fromDict(cls, d: dict[str, tp.Any]):
        return cls(**d)


CI = tp.TypeVar('CI', bound=GenericCollectionDictItem)  # collection item type


@attrs.define(slots=False)
class GenericCollection(ABC, tp.Generic[K, CI]): # (minor note: it would be helpful to specify CI[K] here but python syntax doesn't yet allow for this)
    """
    Base class to implement collection behavior and signaling for various session model components
    """

    _items: dict[K, CI] = attrs.field(factory=dict)

    sigItemsAboutToChange: Signal[list[K], list[str] | None] = attrs.field(init=False, eq=False, factory=Signal, repr=False)
    """
    This signal includes list of keys of collection items about to change, and optionally a list of 
    keys of attributes about to change;  if second arg is None, all attributes should be assumed to
    be about to change.
    """

    sigItemsChanged: Signal[list[K], list[str] | None] = attrs.field(init=False, eq=False, factory=Signal, repr=False)
    """
    This signal includes list of keys of changed collection items, and optionally a list of keys of
    changed attributes; if second arg is None, all attributes should be assumed to have changed.
    """

    sigItemKeyAboutToChange: Signal = attrs.field(init=False, eq=False, factory=lambda: Signal((K, K)), repr=False)
    sigItemKeyChanged: Signal = attrs.field(init=False, eq=False, factory=lambda: Signal((K, K)), repr=False)
    """
    Emitted in addition to sigItemsAboutToChange and sigItemsChanged specifically when an item key changes.

    This is because when a key does change, everything else about a item may have changed, so these other signals don't include a list of attributes indicating the source of the change.
    """

    def __attrs_post_init__(self):
        for key, item in self._items.items():
            assert item.key == key
            item.sigItemAboutToChange.connect(self._onItemAboutToChange)
            item.sigItemChanged.connect(self._onItemChanged)
            item.sigKeyAboutToChange.connect(self._onItemKeyAboutToChange)
            item.sigKeyChanged.connect(self._onItemKeyChanged)

    def addItem(self, item: CI):
        assert item.key not in self._items
        return self.setItem(item=item)

    def deleteItem(self, key: K):
        self.deleteItems([key])

    def deleteItems(self, keys: list[K]):
        assert all(key in self._items for key in keys)
        logger.info(f'Deleting {keys}')

        self.sigItemsAboutToChange.emit(keys, None)
        for key in keys:
            self._items[key].sigItemAboutToChange.disconnect(self._onItemAboutToChange)
            self._items[key].sigKeyAboutToChange.disconnect(self._onItemKeyAboutToChange)
            self._items[key].sigKeyChanged.disconnect(self._onItemKeyChanged)
            self._items[key].sigItemChanged.disconnect(self._onItemChanged)

            del self._items[key]

        self.sigItemsChanged.emit(keys, None)

    def setItem(self, item: CI):
        self.sigItemsAboutToChange.emit([item.key], None)
        if item.key in self._items:
            self._items[item.key].sigItemAboutToChange.disconnect(self._onItemAboutToChange)
            self._items[item.key].sigKeyAboutToChange.disconnect(self._onItemKeyAboutToChange)
            self._items[item.key].sigKeyChanged.disconnect(self._onItemKeyChanged)
            self._items[item.key].sigItemChanged.disconnect(self._onItemChanged)
        self._items[item.key] = item

        item.sigItemAboutToChange.connect(self._onItemAboutToChange)
        item.sigKeyAboutToChange.connect(self._onItemKeyAboutToChange)
        item.sigKeyChanged.connect(self._onItemKeyChanged)
        item.sigItemChanged.connect(self._onItemChanged)

        self.sigItemsChanged.emit([item.key], None)

    def setItems(self, items: list[CI]):
        # assume all keys are changing, though we could do comparisons to find subset changed
        # note that this can also be used for reordering items
        oldKeys = list(self.keys())
        newKeys = [item.key for item in items]
        combinedKeys = list(set(oldKeys) | set(newKeys))
        self.sigItemsAboutToChange.emit(combinedKeys, None)
        for key in oldKeys:
            self._items[key].sigItemAboutToChange.disconnect(self._onItemAboutToChange)
            self._items[key].sigKeyAboutToChange.disconnect(self._onItemKeyAboutToChange)
            self._items[key].sigKeyChanged.disconnect(self._onItemKeyChanged)
            self._items[key].sigItemChanged.disconnect(self._onItemChanged)
        self._items = {item.key: item for item in items}
        for key, item in self._items.items():
            self._items[key].sigItemAboutToChange.connect(self._onItemAboutToChange)
            self._items[key].sigKeyAboutToChange.connect(self._onItemKeyAboutToChange)
            self._items[key].sigKeyChanged.connect(self._onItemKeyChanged)
            self._items[key].sigItemChanged.connect(self._onItemChanged)
        self.sigItemsChanged.emit(combinedKeys, None)

    def setAttribForItems(self, keys: Sequence[K], attribsAndValues: dict[str, Sequence[tp.Any]]) -> None:
        """
        Change an attribute for multiple items at once, without signaling separately for each item.
        E.g. for changing visibility of multiple targets simultaneously.

        :param keys: Identifying keys of items to change
        :param attribsAndValues: mapping of attribute -> list of new values, where this list is of the same length as specified keys.
        :return: None
        """
        assert 'key' not in attribsAndValues, 'Should change key separately, not via setAttribForItems'
        for attrib, values in attribsAndValues.items():
            assert len(values) == len(keys)

        changingKeys = list()
        changingAttribsAndValues = dict()
        for iKey, key in enumerate(keys):
            for attrib, values in attribsAndValues.items():
                assert len(values) == len(keys)
                if not array_equalish(getattr(self._items[key], attrib), values[iKey]):
                    changingKeys.append(key)
                    if attrib not in changingAttribsAndValues:
                        changingAttribsAndValues[attrib] = list()
                    changingAttribsAndValues[attrib].append(values[iKey])
                    break

        if len(changingKeys) == 0:
            return

        self.sigItemsAboutToChange.emit(changingKeys, list(changingAttribsAndValues.keys()))
        with self.sigItemsAboutToChange.blocked(), self.sigItemsChanged.blocked():
            for iKey, key in enumerate(changingKeys):
                for attrib, values in changingAttribsAndValues.items():
                    setattr(self._items[key], attrib, values[iKey])
        self.sigItemsChanged.emit(changingKeys, list(changingAttribsAndValues.keys()))

    def _onItemAboutToChange(self, key: str, attribKeys: tp.Optional[list[str]] = None):
        self.sigItemsAboutToChange.emit([key], attribKeys)

    def _onItemKeyAboutToChange(self, fromKey: str, toKey: str):
        assert toKey not in self._items
        self.sigItemKeyAboutToChange.emit(fromKey, toKey)
        self.sigItemsAboutToChange.emit([fromKey, toKey], None)

    def _onItemKeyChanged(self, fromKey: str, toKey: str):
        assert toKey not in self._items
        self._items = {(toKey if key == fromKey else key): val for key, val in self._items.items()}
        self.sigItemKeyChanged.emit(fromKey, toKey)
        self.sigItemsChanged.emit([fromKey, toKey], None)

    def _onItemChanged(self, key: str, attribKeys: tp.Optional[list[str]] = None):
        self.sigItemsChanged.emit([key], attribKeys)

    def __getitem__(self, key):
        return self._items[key]

    def __setitem__(self, key, item: CI):
        assert key == item.key
        self.setItem(item=item)

    def __iter__(self):
        return iter(self._items)

    def __len__(self):
        return len(self._items)

    def keys(self):
        return self._items.keys()

    def items(self) -> collections.ItemsView[K, CI]:
        return self._items.items()

    def get(self, *args, **kwargs):
        return self._items.get(*args, **kwargs)

    def values(self):
        return self._items.values()

    def merge(self: C, otherItems: C):
        self.sigItemsAboutToChange.emit(list(otherItems.keys()), None)

        with self.sigItemsAboutToChange.blocked(), self.sigItemsChanged.blocked():
            for item in otherItems.values():
                self.setItem(item)

        self.sigItemsChanged.emit(list(otherItems.keys()), None)

    def asList(self) -> list[dict[str, tp.Any]]:
        return [item.asDict() for item in self._items.values()]

    @classmethod
    def fromList(cls: tp.Type[C], itemList: list[dict[K, tp.Any]]) -> C:
        raise NotImplementedError('Should be implemented by subclass')



C = tp.TypeVar('C', bound=GenericCollection)  # collection type



import attrs
import logging
import typing as tp
from qtpy import QtGui

from NaviNIBS.Navigator.GUI.CollectionModels import CollectionTableModel, K
from NaviNIBS.Navigator.Model.Triggering import Hotkey, Hotkeys, HotkeyTriggerSource
from NaviNIBS.util import makeStrUnique

logger = logging.getLogger()


@attrs.define(slots=False, kw_only=True)
class HotkeysTableModel(CollectionTableModel[str, Hotkeys, Hotkey]):
    _triggerSourceKey: str
    _collection: Hotkeys = attrs.field(init=False, repr=False)

    _hasPlaceholderNewRow: bool = True
    _placeholderNewRowDefaults: dict[str, tp.Any] = attrs.field(factory=lambda: dict(key='<NewHotkey>'))

    def __attrs_post_init__(self):
        triggerSource = self._session.triggerSources[self._triggerSourceKey]
        assert isinstance(triggerSource, HotkeyTriggerSource)
        self._collection = triggerSource.hotkeys

        # TODO: subscribe to changes in triggerSource

        self._attrColumns = [
            'key',
            'action',
            'keyboardDeviceID'
        ]

        self._editableColumns = self._attrColumns.copy()

        self._collection.sigItemsAboutToChange.connect(self._onCollectionAboutToChange)
        self._collection.sigItemsChanged.connect(self._onCollectionChanged)

        self._addNewRowFromEditedPlaceholder = self.__addNewRowFromEditedPlaceholder

        super().__attrs_post_init__()

    def __addNewRowFromEditedPlaceholder(self, **kwargs) -> str:
        if 'key' not in kwargs:
            kwargs['key'] = makeStrUnique('Hotkey', self._collection.keys())
        if 'action' not in kwargs:
            kwargs['action'] = 'sample'
        item = Hotkey(**kwargs)
        self._collection.addItem(item)
        return item.key


import attrs
import numpy as np
import typing as tp

from NaviNIBS.Navigator.GUI.CollectionModels import CollectionTableModel, K, logger
from NaviNIBS.Navigator.Model.SubjectRegistration import HeadPoint, HeadPoints
from NaviNIBS.util.Transforms import applyTransform
from NaviNIBS.util.pyvista.dataset import find_closest_point


@attrs.define(slots=False)
class HeadPointsTableModel(CollectionTableModel[int, HeadPoints, HeadPoint]):
    _collection: HeadPoints = attrs.field(init=False, repr=False)

    def __attrs_post_init__(self):
        self._collection = self._session.subjectRegistration.sampledHeadPoints

        self._columns = [
            'distFromSkin',
            'XYZ'
        ]
        self._derivedColumns = dict(
            distFromSkin=self._getDistFromSkinForIndex,
            XYZ=self._getXYZForIndex
        )
        self._columnLabels = dict(
            distFromSkin='Dist from skin (mm)',
            XYZ='XYZ (tracker rel.)'
        )

        self._collection.sigHeadpointsAboutToChange.connect(self._onCollectionAboutToChange)
        self._collection.sigHeadpointsChanged.connect(self._onCollectionChanged)

        # subscribe to changes that will affect derived columns
        self._session.subjectRegistration.sigTrackerToMRITransfAboutToChange.connect(
            lambda: self._onCollectionAboutToChange(None, ['distFromSkin',]))
        self._session.subjectRegistration.sigTrackerToMRITransfChanged.connect(
            lambda: self._onCollectionChanged(None, ['distFromSkin',]))

        super().__attrs_post_init__()

    def _getDistFromSkinForIndex(self, index: int) -> str:
        pt = self._collection[index]
        if self._session.subjectRegistration.trackerToMRITransf is None:
            return ''
        pt_MRISpace = applyTransform(self._session.subjectRegistration.trackerToMRITransf, pt, doCheck=False)
        closestPtIndex = find_closest_point(self._session.headModel.skinSurf, pt_MRISpace)
        closestPt = self._session.headModel.skinSurf.points[closestPtIndex, :]
        dist = np.linalg.norm(closestPt - pt_MRISpace)
        # TODO: improvie dist estimation by allowing for nearest point to be between mesh vertices
        return '%.2f' % dist

    def _getXYZForIndex(self, index: int) -> str:
        pt = self._collection[index]
        ptStr = ','.join(['%.1f' % val for val in pt])
        return ptStr


from qtpy import QtWidgets, QtGui, QtCore


class QWidgetWithCloseSignal(QtWidgets.QWidget):

    sigAboutToClose = QtCore.Signal()

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        self.sigAboutToClose.emit()
        return super().closeEvent(event=event)


class QMainWindowWithCloseSignal(QtWidgets.QMainWindow):

    sigAboutToClose = QtCore.Signal()

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        self.sigAboutToClose.emit()
        return super().closeEvent(event)


from __future__ import annotations
import attrs
import logging
import nibabel as nib
import nitransforms as nit
import numpy as np
import os
import typing as tp
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from NaviNIBS.Navigator.Model.Session import Session
from NaviNIBS.util.attrs import attrsAsDict
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.Transforms import applyTransform, invertTransform

from NaviNIBS.Navigator.Model.GenericCollection import GenericCollection, GenericCollectionDictItem

logger = logging.getLogger(__name__)


@attrs.define
class CoordinateSystem(GenericCollectionDictItem[str]):
    _description: str = ''
    _isVisible: bool = True
    _isAutogenerated: bool = False
    _session: tp.Optional[Session] = attrs.field(default=None, repr=False)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @property
    def description(self):
        return self._description

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, newSes: tp.Optional[Session]):
        self.sigItemAboutToChange.emit(self._key)
        self._session = newSes
        self.sigItemChanged.emit(self._key)

    @property
    def isVisible(self):
        return self._isVisible

    @property
    def isAutogenerated(self):
        return self._isAutogenerated

    def transformFromWorldToThis(self, coords: np.ndarray) -> np.ndarray:
        raise NotImplementedError  # should be implemented by subclass

    def transformFromThisToWorld(self, coords: np.ndarray) -> np.ndarray:
        raise NotImplementedError  # should be implemented by subclass

    def asDict(self):
        return attrsAsDict(self, exclude=['session'])


@attrs.define(kw_only=True)
class AffineTransformedCoordinateSystem(CoordinateSystem):
    _transfThisToWorld: tp.Optional[np.ndarray] = None
    _transfWorldToThis: tp.Optional[np.ndarray] = None

    __transfWorldToThis: tp.Optional[np.ndarray] = attrs.field(init=False, default=None)
    __transfThisToWorld: tp.Optional[np.ndarray] = attrs.field(init=False, default=None)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.sigItemChanged.connect(lambda *args, **kwargs: self.clearCache())

        assert (self._transfThisToWorld is None) != (self._transfWorldToThis is None),\
            'Must specify transfThisToWorld or transfWorldToThis'

    @property
    def transfThisToWorld(self):
        if self._transfThisToWorld is not None:
            return self._transfThisToWorld
        else:
            if self.__transfThisToWorld is None:
                self.__transfThisToWorld = invertTransform(self._transfWorldToThis)
            return self.__transfThisToWorld

    @property
    def transfWorldToThis(self):
        if self._transfWorldToThis is not None:
            return self._transfWorldToThis
        else:
            if self.__transfWorldToThis is None:
                self.__transfWorldToThis = invertTransform(self._transfThisToWorld)
            return self.__transfWorldToThis

    def clearCache(self):
        self.__transfThisToWorld = None
        self.__transfWorldToThis = None

    def transformFromWorldToThis(self, coords: np.ndarray) -> np.ndarray:
        return applyTransform(self.transfWorldToThis, coords, doCheck=False)

    def transformFromThisToWorld(self, coords: np.ndarray) -> np.ndarray:
        return applyTransform(self.transfThisToWorld, coords, doCheck=False)


@attrs.define(kw_only=True)
class NonlinearTransformedCoordinateSystem(CoordinateSystem):
    _deformationFieldThisToWorld_filepath: str = None
    _deformationFieldWorldToThis_filepath: str = None

    _deformationFieldThisToWorld: tp.Optional[nib.Nifti1Image] = attrs.field(init=False, default=None)
    _deformationFieldWorldToThis: tp.Optional[nib.Nifti1Image] = attrs.field(init=False, default=None)

    _transfThisToWorld: tp.Optional[nit.nonlinear.DenseFieldTransform] = attrs.field(init=False, default=None)
    _transfWorldToThis: tp.Optional[nit.nonlinear.DenseFieldTransform] = attrs.field(init=False, default=None)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.sigItemChanged.connect(lambda *args, **kwargs: self.clearCache())

    @property
    def deformationFieldThisToWorld(self):
        if self._deformationFieldThisToWorld is None:
            logger.info(f'Loading deformation field from {self._deformationFieldThisToWorld_filepath}')
            self._deformationFieldThisToWorld = nib.load(self._deformationFieldThisToWorld_filepath)
        return self._deformationFieldThisToWorld

    @property
    def deformationFieldWorldToThis(self):
        if self._deformationFieldWorldToThis is None:
            logger.info(f'Loading deformation field from {self._deformationFieldWorldToThis_filepath}')
            self._deformationFieldWorldToThis = nib.load(self._deformationFieldWorldToThis_filepath)
        return self._deformationFieldWorldToThis

    @property
    def transfThisToWorld(self):
        if self._transfThisToWorld is None:
            self._transfThisToWorld = nit.nonlinear.DenseFieldTransform(
                field=self.deformationFieldThisToWorld,
                is_deltas=False)  # TODO: determine whether also need to specify reference arg
        return self._transfThisToWorld

    @property
    def transfWorldToThis(self):
        if self._transfWorldToThis is None:
            self._transfWorldToThis = nit.nonlinear.DenseFieldTransform(
                field=self.deformationFieldWorldToThis,
                is_deltas=False)  # TODO: determine whether also need to specify reference arg
        return self._transfWorldToThis

    def clearCache(self):
        self._deformationFieldThisToWorld = None
        self._deformationFieldWorldToThis = None
        self._transfThisToWorld = None
        self._transfWorldToThis = None

    def transformFromWorldToThis(self, coords: np.ndarray) -> np.ndarray:
        res = self.transfWorldToThis.map(coords)
        if res.ndim > coords.ndim:
            assert coords.ndim == 1
            res = res.flatten()
        return res

    def transformFromThisToWorld(self, coords: np.ndarray) -> np.ndarray:
        res = self.transfThisToWorld.map(coords)
        if res.ndim > coords.ndim:
            assert coords.ndim == 1
            res = res.flatten()
        return res


@attrs.define
class CoordinateSystems(GenericCollection[str, CoordinateSystem]):
    _session: tp.Optional[Session] = attrs.field(default=None, repr=False)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        if self.session is not None:
            self.session.headModel.sigFilepathChanged.connect(self.autogenerateCoordinateSystems)
            self.autogenerateCoordinateSystems()

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, newSes: tp.Optional[Session]):
        if self._session is newSes:
            return
        if self._session is not None:
            self.session.headModel.sigFilepathChanged.disconnect(self.autogenerateCoordinateSystems)
        self.setAttribForItems(self.keys(), dict(session=[newSes for key in self.keys()]))
        self._session = newSes
        if newSes is not None:
            newSes.headModel.sigFilepathChanged.connect(self.autogenerateCoordinateSystems)
        self.autogenerateCoordinateSystems()

    def autogenerateCoordinateSystems(self):
        if self.session is not None:
            if self.session.headModel.filepath is not None:
                # autogenerate from files generated by SimNIBS
                logger.info('Autogenerating coord system transforms from SimNIBS results')

                parentDir = os.path.dirname(self.session.headModel.filepath)  # simnibs results dir
                subStr = os.path.splitext(os.path.basename(self.session.headModel.filepath))[0]  # e.g. 'sub-1234'
                m2mDir = os.path.join(parentDir, 'm2m_' + subStr)

                inputPath = os.path.join(m2mDir, 'toMNI', 'MNI2conform_12DOF.txt')
                if os.path.exists(inputPath):
                    with open(inputPath, 'r') as f:
                        MNIToMRITransf = np.loadtxt(f)

                    coordSys = AffineTransformedCoordinateSystem(
                        key='MNI_SimNIBS12DoF',
                        description='SimNIBS-generated 12DoF MNI transform',
                        transfThisToWorld=MNIToMRITransf,
                        isAutogenerated=True
                    )
                    if coordSys.key in self._items:
                        assert self._items[coordSys.key].isAutogenerated
                    self.setItem(coordSys)
                else:
                    logger.error(f'Did not find expected input file at {inputPath}')

                if False:  # TODO: debug, re-enable after addressing performance issues with nonlinear transform mapping
                    inputPath = os.path.join(m2mDir, 'toMNI', 'MNI2Conform_nonl.nii')
                    if os.path.exists(inputPath):
                        inputPath2 = os.path.join(m2mDir, 'toMNI', 'Conform2MNI_nonl.nii')
                        if os.path.exists(inputPath2):
                            coordSys = NonlinearTransformedCoordinateSystem(
                                key='MNI_SimNIBSNonlinear',
                                description='SimNIBS-generated nonlinear MNI transform',
                                deformationFieldThisToWorld_filepath=inputPath,
                                deformationFieldWorldToThis_filepath=inputPath2,
                                isAutogenerated=True
                            )
                            assert coordSys.key not in self._items
                            self.setItem(coordSys)
                        else:
                            logger.error(f'Did not find expected input file at {inputPath2}')
                    else:
                        logger.error(f'Did not find expected input file at {inputPath}')


    def asList(self) -> list[dict[str, tp.Any]]:
        # (don't include auto-generated coord systems since they can be restored separately)
        return [coordSys.asDict() for coordSys in self._items.values() if not coordSys.isAutogenerated]

    @classmethod
    def fromList(cls, coordSysList: list[dict[str, tp.Any]]) -> CoordinateSystems:
        coordinateSystems = {}
        for coordSysDict in coordSysList:
            match coordSysDict['__type']:
                case 'AffineTransformedCoordinateSystem':
                    CoordSysCls = AffineTransformedCoordinateSystem
                case 'NonlinearTransformedCoordinateSystem':
                    CoordSysCls = NonlinearTransformedCoordinateSystem
                case _:
                    raise NotImplementedError("Unexpected CoordinateSystem type: {coordSysDict['__type']}")

            coordinateSystems[coordSysDict['key']] = CoordSysCls.fromDict(coordSysDict)

        return cls(items=coordinateSystems)



import attr
import asyncio
import logging
from PySide6 import QtWidgets, QtGui, QtCore
import pyqtgraph as pg
import typing as tp

from .QWidgetWithCloseSignal import QMainWindowWithCloseSignal
from ..Signaler import Signal


from NaviNIBS.util import exceptionToStr
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError

logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True, eq=False, kw_only=True)
class RunnableAsApp:
    _appName: tp.Optional[str] = None
    _appAsyncPollPeriod = 0.01
    _doRunAsApp: bool = False
    _appLogEveryNLoops: tp.Optional[int] = None
    _theme: str = 'auto'  # auto, light, or dark

    _prevSetTheme: str = attr.ib(init=False, default=None)
    _app: QtWidgets.QApplication = attr.ib(init=False)
    _appIconPath: str | None = attr.ib(init=False, default=None)
    _Win: tp.Callable[..., QMainWindowWithCloseSignal] = attr.ib(default=QMainWindowWithCloseSignal)
    """
    Can specify Win to point to a different MainWindow class (e.g. for docking support)
    """
    _win: QMainWindowWithCloseSignal = attr.ib(init=False, default=None)
    _appIsClosing: bool = attr.ib(init=False, default=False)

    def __attrs_post_init__(self):
        if self._doRunAsApp:
            if self._appName is None:
                self._appName = self.__class__.__name__
            logger.info('Initializing %s' % (self._appName,))

            logger.debug('Initializing GUI window')
            self._app = pg.mkQApp(self._appName)

            if self._appIconPath is not None:

                import platform
                isWin = platform.system() == 'Windows'
                if isWin:
                    # workaround to show icon in Windows task bar
                    # based on https://stackoverflow.com/questions/1551605/how-to-set-applications-taskbar-icon-in-windows-7/1552105#1552105
                    import ctypes
                    myappid = f"NaviNIBS.{self._appName.replace(' ', '')}.subproduct.version"  # arbitrary string
                    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

                self._app.setWindowIcon(QtGui.QIcon(self._appIconPath))

            theme = self._theme
            self._theme = None
            self.theme = theme  # trigger setter

            self._win = self._Win()
            self._win.setWindowTitle(self._appName)
            self._win.sigAboutToClose.connect(self._onAppAboutToQuit)

    @property
    def theme(self):
        return self._theme

    @theme.setter
    def theme(self, theme):
        if self._theme == theme:
            return
        self._theme = theme
        if theme.lower() == 'auto':
            import darkdetect
            if darkdetect.isDark():
                theme = 'dark'
            else:
                theme = 'light'
        import qtawesome as qta
        match theme.lower():
            case 'light':
                if self._prevSetTheme is None or self._prevSetTheme == 'light':
                    # no change needed, avoid calling qta.light since it also changes app style
                    pass
                else:
                    qta.light(self._app)
            case 'dark':
                qta.dark(self._app)
            case _:
                raise ValueError(f'Unknown theme: {theme}')

        self._prevSetTheme = theme.lower()

    def _onAppAboutToQuit(self):
        logger.info('About to quit')
        self._appIsClosing = True

    @property
    def appIsClosing(self):
        return self._appIsClosing

    async def _runLoop(self):
        logger.debug('Running %s' % (self.__class__.__name__,))
        if self._appLogEveryNLoops is None:
            counter = 1
        else:
            counter = 0
        while not self._appIsClosing:
            if counter == 0:
                logger.debug('Main loop: process Qt events')
            try:
                self._app.processEvents()
            except Exception as e:
                logger.error(f'Exception while processing Qt events: {exceptionToStr(e)}')
                raise e
            if counter == 0:
                logger.debug('Main loop: async')
            await asyncio.sleep(self._appAsyncPollPeriod)
            if counter == 0:
                logger.debug('Main loop: end')
            # b/c doing QT event processing ourselves, for some reason the aboutToQuit signal is never emitted,
            # so check (infrequently) whether window has been closed and quit manually
            if self._appLogEveryNLoops is not None:
                counter = (counter + 1) % self._appLogEveryNLoops

    @classmethod
    def createAndRunAsTask(cls, *args, **kwargs):
        logger.debug('Creating %s' % (cls.__name__,))
        self = cls(*args, doRunAsApp=True, **kwargs)
        asyncio.create_task(asyncTryAndLogExceptionOnError(self._runLoop))
        return self

    @classmethod
    async def createAndRun_async(cls, *args, **kwargs):
        logger.debug('Creating %s' % (cls.__name__,))
        try:
            self = cls(*args, doRunAsApp=True, **kwargs)
        except Exception as e:
            raise e

        await self._runLoop()

    @classmethod
    def createAndRun(cls, *args, **kwargs):
        if True:
            from ..AsyncRunner import asyncioRunAndHandleExceptions
            asyncioRunAndHandleExceptions(cls.createAndRun_async, *args, **kwargs)
        else:
            asyncio.run(cls.createAndRun_async(*args, **kwargs), debug=True)


@attr.s(auto_attribs=True, eq=False)
class TestGUI(RunnableAsApp):
    _appName: str = 'Test GUI'

    def __attrs_post_init__(self):
        logger.info('Initializing TestGUI')

        super().__attrs_post_init__()

        self._wdgt = QtWidgets.QWidget()

        if self._doRunAsApp:
            self._win.setCentralWidget(self._wdgt)
            logger.info('Showing window')
            self._win.show()

    def _onAppAboutToQuit(self):
        self.cleanup()
        super()._onAppAboutToQuit()

    def cleanup(self):
        logger.info('Cleanup')


if __name__ == '__main__':
    TestGUI.createAndRun()


import logging
logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s.%(msecs)03d  %(process)6d %(filename)20s %(lineno)4d %(levelname)5s: %(message)s',
                    datefmt='%H:%M:%S')

__version__ = '0.2.3'
